import unittest
from tfv.extractor import FvExtractor
import numpy as np
import re
from pathlib import Path
import pandas as pd

path = Path(__file__).parent / 'data'
xtr = FvExtractor(path / 'HYD_002_mini.nc')

class TestFvExtractor(unittest.TestCase):
    def test_get_sheet_cell_V_defaults(self):
        var = 'V' # magic var
        time = 1
        
        data = xtr.get_sheet_cell(var, time)
        
        self.assertEqual(data.shape[0], xtr.nc2)
    
    def test_get_sheet_cell_TEMP_dave(self):
        var = 'TEMP'
        time = 0
        
        data = xtr.get_sheet_cell(var, time, datum='depth', limits=(0,2.1))      
        
        self.assertEqual(data.shape[0], xtr.nc2)
        
    def test_get_sheet_node_SAL_agg(self):
        var = 'SAL' 
        time = 0
        
        data = xtr.get_sheet_node(var, time, agg='max')
        
        self.assertEqual(data.shape[0], xtr.nv2)
        
    def test_get_sheet_grid(self):
        var = 'V' # magic var
        time = 2
        
        bbox = '159.07645908,-31.40185855,159.11088280,-31.37972153'
        xl, yl, xu, yu = [float(x) for x in bbox.split(',')]
        xg = np.linspace(xl, xu, 19)
        yg = np.linspace(yl, yu, 21)
        
        data = xtr.get_sheet_grid(var, time, xg, yg, datum='height', agg='max')
        
        self.assertEqual(data.shape, (21,19))
        
    def test_get_curtain_grid(self):
        ''' covers pretty much all the curtain functions - if it works, they work '''
        var = 'V' # magic var
        time = '2011-02-01 03:00:43.868682'
        
        with open(path / 'polyline.txt', 'r') as f:
            ls = f.readline()
        
        polyline = np.asarray([float(x) for x in re.findall('-?\d+.\d+', ls)])
        polyline = polyline.reshape([polyline.shape[0]//2, -1])
        
        xg = np.linspace(0, 2400, 20)  # Rough length of polyline
        zg = np.linspace(0, 10, 5)  # Rough depth? 
        
        data = xtr.get_curtain_grid(var, time, polyline, xg, -zg)
        
        data_check = data.sum() > 0
        dim_check = data.shape == (5,20)
        self.assertTrue(all([data_check, dim_check]))
        
    def test_get_profile_cell(self):
        ''' legacy function, that DOUBLES each pt (to create discrete "cells")'''
        
        var = 'V_y'
        time = pd.Timestamp('2011-02-01 02:00:00.263494800')
        lat, lon = -31.37719201, 159.08970821
        
        ii = xtr._timestep_index(time)
        idx = xtr.get_cell_index(lon, lat)
        idx_lfz = xtr.idx4 == idx
        lfz = xtr.get_z_layer_faces(ii)[idx_lfz]
        ndims = int((len(lfz) - 1) * 2)
        
        data = xtr.get_profile_cell(var, time, (lon, lat))
        
        self.assertEqual(data.shape[0], ndims)

if __name__ == '__main__':
    unittest.main()



"""A module for reading and writing TUFLOW FV restart files"""

import os
import numpy as np
from netCDF4 import Dataset

def write_restart_file(nc2, nc3, time_stamp, cell_Zb, fv_data, out_file):
    """
    Creates a new restart file from data.

    Parameters
    ----------
    nc2 : int
        Number of 2d cells.
    nc3 : int
        Number of 3d cells.
    time_stamp : float
        Time as TUFLOW FV time stamp (hours from 1/1/1990).
    cell_Zb : ndarray
        Cell elevation as (nc2,) array.
    fv_data : ndarray
        Conserved 3D variables as (nc3, n) array (depth, V_x, V_y, SAL, TEMP, SED_1, .... , SED_N)
    out_file : str
        Output path of restart file.
    """

    # get some basic parameters
    t = np.array(time_stamp * 3600)  # time in seconds since 01/01/1990 00:00:00
    nv = fv_data.shape[1]  # number of conserved variables
    dims = np.array([nc2, nc3, nv])  # array of dimensions

    # scale data by depth and transpose
    fv_data[:, 1:] = fv_data[:, 1:]*fv_data[:, 0:1]
    fv_data = fv_data.transpose()

    # write the data in binary format as shown
    with open(out_file, 'wb') as f:
        t.astype(np.float64).tofile(f)
        dims.astype(np.int32).tofile(f)
        cell_Zb.astype(np.float32).tofile(f)
        fv_data.transpose().astype(np.float32).tofile(f)

def read_restart_file(restart_file):
    """Reads data from a TUFLOW FV restart file"""
    with open(restart_file, 'rb') as f:
        time_stamp = np.fromfile(f, np.float64, 1)[0]/3600
        nc2 = np.fromfile(f, np.int32, 1)[0]
        nc3 = np.fromfile(f, np.int32, 1)[0]
        nv = np.fromfile(f, np.int32, 1)[0]

        cell_Zb = np.fromfile(f, np.float32, nc2)

        fv_data = np.zeros((nv, nc3,), np.float32)
        for aa in range(nv):
            fv_data[aa, :] = np.fromfile(f, np.float32, nc3)
        fv_data = fv_data.transpose()

        good, bad = fv_data[:, 0] != 0, fv_data[:, 0] == 0
        fv_data[good, 1:] = fv_data[good, 1:] / fv_data[good, 0:1]
        fv_data[bad, :] = np.nan

    return nc2, nc3, time_stamp, cell_Zb, fv_data

def restart_from_result(old_result_file, new_result_file, out_file, n_sigma, time_stamp, variables):
    """
    Creates a restart file from an existing TUFLOW FV result file, using a separate result file for geometry input.

    Parameters
    ----------
    old_result_file : str
        Path of result file from which restart is being generated.
    new_result_file : str
        Path of result file from which to use geometry data (mesh and cell elevations).
    out_file : str
        Output path of restart file.
    n_sigma : int
        Number of sigma layers being used in new model.
    time_stamp : float
        Time as TUFLOW FV time stamp (hours from 1/1/1990).
    variables : list
        List of conserved variables to include in the restart file (V_x, V_y, SAL, TEMP, SED_1, ...).
    """

    # get old result file netCDF handle
    old = Dataset(old_result_file)

    # get new result file netCDF handle
    new = Dataset(new_result_file)

    # get time index from target result file
    tt = np.argmin(np.abs(old['ResTime'][:] - time_stamp))

    # find the nearest 3D index for each cell
    index3D = np.array([], dtype=np.int32)  # maps new 3D to old 3D
    for aa in range(new.dimensions['NumCells2D'].size):

        # get the distance to each 2D cell
        dx = old['cell_X'] - new['cell_X'][aa]
        dy = old['cell_Y'] - new['cell_Y'][aa]
        distance = np.hypot(dx, dy)

        # get the index of the nearest old 2D cell
        nearest2D = np.argmin(distance)

        # get the new water level and bed level
        wl, bl = old['H'][tt, nearest2D].data, new['cell_Zb'][aa].data

        # check if old and new cell is\isn't dry
        dryDepth = new.getncattr('Dry depth')
        newDry = (wl - bl) < dryDepth
        oldDry = (old['stat'][tt, nearest2D] == 0)

        # if new cell is wet but old cell is dry, remap, otherwise bad data
        if (not newDry) and oldDry:
            distance[old['stat'][tt, :] == 0] = np.inf
            nearest2D = np.argmin(distance)
            wl = old['H'][tt, nearest2D].data

        # get the NEW layer face Z for current cell
        idx3 = new['idx3'][aa] - 1
        nlfz = new['NL'][aa] + 1
        idx4 = idx3 + aa

        lfzNew = new['layerface_Z'][tt, idx4:idx4 + nlfz].data

        # update the sigma layers using new water level
        dzTop = (wl - lfzNew[n_sigma]) / n_sigma
        lfzNew[0:n_sigma] = wl - dzTop * np.arange(n_sigma)

        # get the OLD layer face Z for current cell
        idx3 = old['idx3'][nearest2D] - 1
        nlfz = old['NL'][nearest2D] + 1
        idx4 = idx3 + nearest2D

        lfzOld = old['layerface_Z'][tt, idx4:idx4 + nlfz].data

        # get the centres to do a minimum distance search
        zcNew = 0.5 * (lfzNew[:-1] + lfzNew[1:])
        zcOld = 0.5 * (lfzOld[:-1] + lfzOld[1:])

        zcOld = np.tile(zcOld, (zcNew.size, 1)).transpose()

        nearest3D = np.argmin(np.abs(zcOld - zcNew), axis=0)

        index3D = np.hstack((index3D, idx3 + nearest3D))

    # get the 2D depth at each 3D cell (used to scale variables for some reason)
    depth = old['H'][tt, (old['idx2'][index3D] - 1)] - new['cell_Zb'][new['idx2'][:] - 1]

    # create empty array for data
    fvData = np.zeros((len(index3D), len(variables) + 1))

    # always set first column to depth
    fvData[:, 0] = depth

    # fill with other conserved variables
    for aa in range(len(variables)):
        if variables[aa] in old.variables:
            fvData[:, aa + 1] = old[variables[aa]][tt, index3D]

    write_restart_file(new.dimensions['NumCells2D'].size, new.dimensions['NumCells3D'].size,
                       time_stamp, new['cell_Zb'][:].data, fvData, out_file)

def _restart_from_result_beta(old_result_file, new_fvc_file, out_file, time_stamp, variables):
    """
    An untested version in the making which directly reads from a .fvc file the new geometry.

    Limited because it does not read for elevation limits set based on material type or shape file.

    This could one day be improved.
    """

    # set default geometry parameters
    cellFile = None
    numSigma = None
    layerFile = None
    minThick = None

    # read parameters from .fvc file
    with open(new_fvc_file, 'r') as f:
        # get start line
        line = f.readline()
        while line != '':
            if 'cell elevation file' in line:
                cellFile = line.split('==')[-1].strip()
            if 'sigma layers' in line:
                numSigma = int(line.split('==')[-1].strip())
            if 'layer faces' in line:
                layerFile = line.split('==')[-1].strip()
            if 'min bottom layer thickness' in line:
                minThick = float(line.split('==')[-1].strip())

            # read the next line
            line = f.readline()

    # get path to .fvc file folder
    fvcFolder = os.path.split(new_fvc_file)[0]

    # set paths to absolute, assume they are relative in the .fvc file
    cellFile = os.path.abspath(os.path.join(fvcFolder, cellFile))
    layerFile = os.path.abspath(os.path.join(fvcFolder, layerFile))

    # read the cell centred elevations
    data = np.loadtxt(cellFile, skiprows=1, delimiter=',', dtype=np.float64)
    cellX, cellY, cellZb = data[:, 0], data[:, 1], data[:, 2]

    # read the fixed elevation layers
    zLayers = np.loadtxt(layerFile, skiprows=1, delimiter=',', dtype=np.float64)

    # get result file netCDF handle
    old = Dataset(old_result_file)

    # get time index from target result file
    tt = np.argmin(np.abs(old['ResTime'] - time_stamp))

    # find the nearest 3D index for each cell
    index3D = np.array([], dtype=np.int32)  # maps new 3D to old 3D
    idx2 = np.array([], dtype=np.int32)  # maps new 2D to 3D
    for aa in range(cellZb.size):

        # get the distance to each 2D cell
        dx = old['cell_X'] - cellX[aa]
        dy = old['cell_Y'] - cellY[aa]
        distance = np.hypot(dx, dy)

        # get the index of the nearest old 2D cell
        nearest2D = np.argmin(distance)

        # check if old and new cell is\isn't dry
        dryDepth = old.getncattr('Dry depth')
        newDry = (old['H'][tt, nearest2D] - cellZb[aa]) < dryDepth
        oldDry = (old['stat'][tt, nearest2D] == 0)

        # if new cell is wet but old cell is dry, remap
        if (not newDry) and oldDry:
            distance[old['stat'][tt, :] == 0] = np.inf
            nearest2D = np.argmin(distance)

        # get the new water level and bed level
        wl, bl = old['H'][tt, nearest2D].data, cellZb[aa]

        # start from the top with the sigma layers
        dzTop = (wl - zLayers[0]) / numSigma
        lfzNew = wl - dzTop * np.arange(numSigma)

        # add in the elevation layers above the bed
        aboveBed = zLayers > bl
        botThick = zLayers - bl
        goodThick = botThick > minThick

        lgi = aboveBed & goodThick

        lfzNew = np.hstack((lfzNew, zLayers[lgi], bl))

        idx2 = np.hstack((idx2, np.repeat(aa, len(lfzNew) - 1)))

        # use layer faces to map new to old 3D points
        idx3 = old['idx3'][nearest2D] - 1
        nlfz = old['NL'][nearest2D] + 1
        idx4 = idx3 + nearest2D

        lfzOld = old['layerface_Z'][tt, idx4:idx4 + nlfz].data

        zcNew = 0.5 * (lfzNew[:-1] + lfzNew[1:])
        zcOld = 0.5 * (lfzOld[:-1] + lfzOld[1:])

        zcOld = np.tile(zcOld, (zcNew.size, 1)).transpose()

        nearest3D = np.argmin(np.abs(zcOld - zcNew), axis=0)

        index3D = np.hstack((index3D, idx3 + nearest3D))

    # get the 2D depth at each 3D cell (used to scale variables for some reason)
    depth = old['H'][tt, (old['idx2'][index3D] - 1)] - cellZb[idx2]

    # create empty array for data
    fvData = np.zeros((len(index3D), len(variables) + 1))

    # always set first column to depth
    fvData[:, 0] = depth

    # fill with other conserved variables
    for aa in range(len(variables)):
        if variables[aa] in old.variables:
            fvData[:, aa + 1] = old[variables[aa]][tt, index3D]

    write_restart_file(cellZb.size, idx2.size, time_stamp, cellZb, fvData, out_file)




import numpy as np
import pandas as pd

# ToDO:
class Expression:
    @classmethod
    def decorator(cls, function_call):
        def wrapper(*args, **kwargs):
            base = args[0]
            variables = args[1]  # variable(s) should always be first required argument

            post_process = kwargs.pop("post_process", None)

            # Special case for Current speed / Direction. Patched to first check if they are present (tfv.xarray)
            if (
                (variables == "V")
                & (post_process is None)
                & ("V" not in base.variables)
            ):
                variables = ["V_x", "V_y"]
                post_process = lambda x: np.hypot(*x)

            elif (
                (variables == "VDir")
                & (post_process is None)
                & ("VDir" not in base.variables)
            ):
                variables = ["V_y", "V_x"]
                post_process = lambda x: (90 - np.arctan2(*x) * 180 / np.pi) % 360

            if type(variables) == str:
                output = function_call(*args, **kwargs)
            elif type(variables) == list:
                data = []
                for v in variables:
                    arr = function_call(base, v, *args[2:], **kwargs)
                    data.append(arr)
                output = np.ma.stack(data)

            if post_process:
                output = post_process(output.astype(float))

            return output

        return wrapper


def unsupported_decorator(function_call):
    def wrapper(*args):
        name = function_call.__name__
        message = "{} is currently not supported".format(name)
        print(message)

    return wrapper



import unittest
import xarray as xr
import tfv.xarray
import matplotlib.pyplot as plt
from pathlib import Path

path = Path(__file__).parent / 'data'
fv = xr.open_dataset(path / 'HYD_002_mini.nc').tfv

class TestTfvDomain(unittest.TestCase):
    def test_get_sheet_H_numcells2d(self):
        ''' Test that get sheet runs with very simple defaults and a 2D var '''
        dims = fv.get_sheet('H')['H'].shape
        self.assertEqual(dims[1], fv.dims['NumCells2D'])

    def test_get_sheet_V_numcells2d(self):
        ''' Test that get sheet runs with the magic 'V' variable '''
        dims = fv.get_sheet('V')['V'].shape 
        self.assertEqual(dims[1], fv.dims['NumCells2D'])
        
    def test_get_sheet_TEMP_time_integer_slice(self):
        ''' Test that get sheet runs with integer slicing '''
        tslc = slice(1, 3)
        ts = fv._obj.isel(Time=tslc).dims['Time'] # Correct behaviour 
        
        # Check func behaviour
        dims = fv.get_sheet('TEMP', time=tslc)['TEMP'].shape
        self.assertEqual(dims[0], ts)
        
    def test_get_sheet_TEMP_time_single_time_integer(self):
        ''' Test that get sheet runs with single integer time '''        
        # Check func behaviour
        dims = fv.get_sheet('TEMP', time=1)['TEMP'].shape
        self.assertEqual(dims[0], 1)
        
    def test_get_sheet_TEMP_time_single_time_datestr(self):
        ''' Test that get sheet runs with a hard datestr '''        
        # Check func behaviour
        dims = fv.get_sheet('TEMP', time='2011-02-01T02:00:00.263494800')['TEMP'].shape
        self.assertEqual(dims[0], 1)
        
    def test_get_statistics_single_stat(self):
        ''' Test a single statistic '''
        # Check func behaviour
        dims = fv.get_statistics('mean', 'V')['V_mean'].shape
        self.assertEqual(dims[0], 1)
        
    def test_get_statistics_several_stats(self):
        ''' Test several statistics '''        
        # Check func behaviour
        stats = ['mean', 'p95', 'p20', 'sum']
        var = 'SAL'
        
        fv2d = fv.get_statistics(stats, var)
        
        svar = f'{var}_{stats[2]}'
        timecheck = fv2d.dims['Time'] == 1
        varcheck = len(fv2d.variables) == len(stats)
        dimcheck = fv2d[svar].shape[1] == fv.dims['NumCells2D']
        
        self.assertTrue(timecheck & varcheck & dimcheck)
        
    def test_get_timeseries(self):
        ''' Test get timeseries with defaults, multi locations. '''        
        # Check func behaviour
        locs = dict(
            pt1=(159.09380193, -31.39236190),
            pt2=(159.11004448, -31.39781102)
        )
        vars = ('H', 'V', 'V_x', 'SAL')
        
        ts = fv.get_timeseries(vars, locs)
        
        timecheck = ts.dims['Time'] == fv.dims['Time']
        varcheck = len(ts.data_vars) == len(vars)
        dimcheck = ts.dims['Location'] == len(locs)
        
        self.assertTrue(timecheck & varcheck & dimcheck)

    def test_get_timeseries_with_dave_opts(self):
        ''' Test get timeseries with depthaveraging options'''        
        # Check func behaviour
        locs = dict(
            pt1=(159.09380193, -31.39236190),
        )
        vars = ('H', 'SAL')
        datum = 'height'
        limits=(0,2.42)
        agg='max'
        
        ts = fv.get_timeseries(vars, locs, datum=datum, limits=limits, agg=agg)
        
        timecheck = ts.dims['Time'] == fv.dims['Time']
        varcheck = len(ts.data_vars) == len(vars)
        dimcheck = 'Location' not in ts.dims  # Only one location, should not be present!
        
        self.assertTrue(timecheck & varcheck & dimcheck)

    def test_get_profile_variables(self):
        ''' Test get profile a set of variables '''
        locs = dict(
            pt1=(159.09380193, -31.39236190),
        )
        vars = ('H', 'V', 'V_x', 'SAL')

        prof = fv.get_profile(locs, variables=vars)

        # Check all variables made it through
        vars_all = len(prof.data_vars) == len(vars)+5
        vars_in = all([x in prof.data_vars for x in vars])
        
        self.assertTrue(all([vars_all, vars_in]))
        
    def test_get_profile_dims(self):
        ''' Test get profile a set of variables '''
        locs = dict(
            pt1=(159.09380193, -31.39236190),
        )
        vars = ('H', 'V', 'V_x', 'SAL')
        nt = 2
        prof = fv.get_profile(locs, time=slice(nt), variables=vars)
        
        dims = all([
            'N1' in prof.dims, # to match TUFLOW FV output
            'NumLayers' in prof.dims,  # 3D model should have this!
            'NumLayerFaces' in prof.dims, # Lfz as standard
        ])

        shp = all([
            prof.dims['Time'] == nt,
        ])
        self.assertTrue(all([dims, shp]))
    
    def test_plot_defaults_H(self):
        ''' Test whether the default plot picks the correct time, and runs without issue. '''
        var = 'H'
        time = '2011-02-01 02:00'

        fv.plot(var, time)
        ax = plt.gca()

        self.assertEqual(time, ax.get_title())
        
    def test_plot_defaults_V(self):
        ''' Test whether the default plot works with the magic V var, and auto-labels'''
        var = 'V'
        vlbl = 'current speed (m s^-1)' # Auto label should assign this.
        time = '2011-02-01 02:30'

        fv.plot(var, time)
        fig = plt.gcf()
        
        # Loop through children to get colorbar
        for c in fig.get_children():
            if hasattr(c.axes, 'get_label'):
                if c.axes.get_label() == '<colorbar>':
                    lbl = c.axes.get_ylabel()

        self.assertEqual(vlbl, lbl)

if __name__ == '__main__':
    unittest.main()



"""
TFV Accessor module to add functionality to native xarray

A Waterhouse

This module provides "meta" functions to access, analyse and plot TUFLOW FV result files, based around underlying Xarray datasets.
The intention of this module is to replace the use of the individual function calls by providing convienent "one-stop shop" methods.
"""

from inspect import getdoc
from pathlib import Path
import re
import xarray as xr

from tfv.miscellaneous import *
from tfv.extractor import FvExtractor, _strip_dataset
from tfv.timeseries import FvTimeSeries
from tfv.visual import *

from tqdm import tqdm
from typing import Union, Literal
import matplotlib
from matplotlib.ticker import FormatStrFormatter

import warnings


def read_tfv(file: Union[str, Path, list], restype="auto"):
    try:
        ds = FvExtractor(file).ds
    except TypeError:
        print("Data does not appear to be a valid TUFLOW FV spatial netcdf file")
    return ds


def _pipe_call(func):
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return result

    return wrapper


def _check_tfv_type(ds):
    atype = ds.attrs.get("Type", "")

    if (atype == "Cell-centred TUFLOWFV output") | ("NumCells2D" in ds.dims):
        restype = "domain"

    elif atype == "Curtain TUFLOWFV output":
        restype = "curtain"

    elif atype in ["TUFLOWFV profile output", "Profile cell from TUFLOWFV output"]:
        restype = "timeseries"
    else:
        restype = "unknown"

    return restype


@xr.register_dataset_accessor("tfv")
class TfvAccessor:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self.restype = _check_tfv_type(xarray_obj)

        if self.restype == "domain":
            self._tfv = TfvDomain(self._obj)

        elif self.restype == "curtain":
            raise NotImplementedError

        elif self.restype == "timeseries":
            self._tfv = TfvTimeseries(self._obj)

        # Register functions from the appropriate module
        for x in dir(self._tfv):
            if x.startswith("_") is False:
                setattr(self, x, getattr(self._tfv, x))

    def __getattr__(self, item):
        # To prevent breaking in-built funcs (e.g., _repr...)
        if item.startswith("_"):
            raise AttributeError(item)

        result = getattr(self._obj, item)

        # Run function, if applicable
        if callable(result):
            result = _pipe_call(result)

        return result

    def __getitem__(self, item):
        result = self._obj[item]
        return result

    def __setitem__(self, item, value):
        self._obj[item] = value

    def _repr_html_(self):
        return self._tfv._repr_html_()

    def __repr__(self):
        return self._tfv.__repr__()


class TfvBase:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self.geo_coords = []

    def __getitem__(self, item):
        result = self._obj[item]
        return result

    def __setitem__(self, item, value):
        self._obj[item] = value

    def _getvar_(self, variable, skip=[]):
        try:
            # Take first variable if not specified
            if variable is None:
                for k, v in self.variables.items():
                    if k not in skip:
                        variable = k
                        attrs = v
                        break

            if variable == "V":
                attrs = {"long_name": "current speed", "units": "m s^-1"}
            elif variable == "VDir":
                attrs = {"long_name": "current direction", "units": "degN TO"}
            elif (variable == "Zb") | (variable == "cell_Zb"):
                # cell Zb doesn't have time dim, and hence will need to be expanded to work with tfv.plot
                variable = "Zb"
                attrs = self._obj["cell_Zb"].attrs
            else:
                attrs = self.variables[variable]

            if ("long_name" in attrs) & ("units" in attrs):
                label = f"{attrs['long_name']} ({attrs['units']})"
            else:
                label = f"{variable} ( - )"
        except AttributeError:
            print(
                "Specified time is not clear - supply Datetime, str date (in ISO format) or integer"
            )
        return variable, label

    @property
    def dims(self):
        return self._obj.dims

    @property
    def data_vars(self):
        return self._obj.drop_vars(self.geo_coords).data_vars

    @property
    def variables(self):
        return {
            k: v.attrs
            for k, v in self.data_vars.items()
            if k not in ["ResTime", "stat", "layerface_Z"]
        }

    def sel(self, *args, **kwargs):
        """pipe sel back through to the tfv object"""
        subset = self._obj.sel(*args, **kwargs)
        tvar = list(self.data_vars.keys())
        if "Time" not in subset.dims:
            subset[tvar] = subset[tvar].expand_dims("Time")
        return subset.tfv

    def isel(self, *args, **kwargs):
        """pipe sel back through to the tfv object"""
        subset = self._obj.isel(*args, **kwargs)
        tvar = list(self.data_vars.keys())
        if "Time" not in subset.dims:
            subset[tvar] = subset[tvar].expand_dims("Time")
        return subset.tfv

    def to_netcdf(self, *args, **kwargs):
        """Writes NETCDF result"""
        self._obj.to_netcdf(*args, **kwargs)


class TfvDomain(TfvBase):
    """Xarray accessor object for working with TUFLOW FV domain (spatial) netcdf files.

    Extends the functionality of native xarray to add methods that assist with typical analyses.

    To use this, call `.tfv` on an xarray dataset based on a TUFLOW FV domain file.
    """

    def __init__(self, xarray_obj):
        TfvBase.__init__(self, xarray_obj)
        self.xtr = None
        self.__load_tfv_domain()

    def __repr__(self):
        if is_notebook() is False:
            return self._obj.drop_vars(self.geo_coords).__repr__()
        else:
            return "TUFLOW FV domain xarray accessor object"

    def _repr_html_(self):
        from IPython.display import display

        return display(self._obj.drop_vars(self.geo_coords))

    def __convert_coords__(self):
        """Promote tfv geometry to coordinates"""
        ds = self._obj

        # Copy all geo variables into coordinates:
        self.geo = {}
        for dv in ds.data_vars.keys():
            if "Time" not in ds[dv].dims:
                self.geo[dv] = ds[dv].values

        # self._obj = ds.drop([c for c in self.geo.keys()])
        self.geo_coords = [c for c in self.geo.keys()]

    def __load_tfv_domain(self):
        if self.xtr is None:
            try:
                self.xtr = FvExtractor(self._obj)
                self.__convert_coords__()
            except TypeError:
                print(
                    "Data does not appear to be a valid TUFLOW-FV spatial netcdf file"
                )

    def get_sheet(
        self,
        variables: Union[str, list],
        time: Union[str, int, pd.Timestamp, slice] = None,
        datum: Literal["sigma", "height", "depth", "elevation"] = "sigma",
        limits: tuple = (0, 1),
        agg: Literal["min", "mean", "max"] = "mean",
    ):
        """Extract a 2D sheet timeseries

        General purpose method for extracting 2D sheet data from a TfvDomain object. This will handle dimension reduction of 3D variables (by default will average over the depth column, i.e., depth-averaged), as well as handling single or multiple timesteps or slices.

        Args:
            variables ([str, list]): variables to extract. ("V" or "VDir" may be requested if "V_x" and "V_y" are present)
            time ([str, pd.Timestamp, int, slice], optional): time indexer for extraction. Defaults to the entire dataset.
            datum (['sigma', 'height', 'depth', 'elevation'], optional): depth-averaging datum. Defaults to 'sigma'. Choose from `height`, `depth`, `elevation` or `sigma`.
            limits (tuple, optional): depth-averaging limits. Defaults to (0,1).
            agg (['min', 'mean', 'max'], optional): depth-averaging aggregation function. Defaults to 'mean'. Choose from `min`, `mean`, or `max`.

        Returns:
            TfvDomain: new TfvDomain object with the requested timeseries variables

        """
        # Use sliced xarray dataset
        xtr = self.xtr
        dsx = xtr._subset_dataset(time)

        times = dsx["Time"]
        ts = [xtr._timestep_index(t) for t in times.values]
        nt = times.shape[0]
        ns = dsx.dims["NumCells2D"]

        # Check input
        if isinstance(variables, str):
            variables = [variables]

        # Add V/VDir methods
        requested_vars = variables.copy()
        if "V" in variables:
            variables.extend(["V_x", "V_y"])
            variables.pop(variables.index("V"))
        if "VDir" in variables:
            variables.extend(["V_x", "V_y"])
            variables.pop(variables.index("VDir"))
        variables = np.unique(variables).tolist()

        # Init output array
        nv = len(variables)
        array = np.ma.zeros((nv, nt, ns))

        # Check if z_data is required
        depth_int = False
        for v in variables:
            if "NumCells3D" in dsx[v].dims:
                depth_int = True

        c = 0
        z_data = None
        for t in tqdm(
            ts,
            desc="...extracting sheet data",
            colour="green",
            delay=1,
        ):
            # ii = xtr._timestep_index(t)
            if depth_int:
                z_data = xtr.get_integral_data(t, datum, limits)
            array[:, c, :] = xtr.get_sheet_cell(
                variables, t, datum, limits, agg=agg, z_data=z_data
            )
            c += 1

        dso = _strip_dataset(dsx.copy(deep=True))
        dims = ("Time", "NumCells2D")
        for v in requested_vars:
            if v == "V":
                vxi = variables.index("V_x")
                vyi = variables.index("V_y")
                arr = np.hypot(array[vxi], array[vyi])
                dso[v] = (dims, arr)
                dso[v].attrs = {"long_name": "current speed", "units": "m s^-1"}

            elif v == "VDir":
                vxi = variables.index("V_x")
                vyi = variables.index("V_y")
                arr = (90 - np.arctan2(array[vyi], array[vxi]) * 180 / np.pi) % 360
                dso[v] = (dims, arr)
                dso[v].attrs = {"long_name": "current direction", "units": "degN TO"}
            else:
                dso[v] = (dims, array[variables.index(v)])
                dso[v].attrs = dsx[v].attrs.copy()

        dso.attrs = self._obj.attrs

        return dso.tfv

    def get_statistics(
        self,
        stats: list[str],
        variables: Union[list, str],
        time: slice = None,
        datum: str = "sigma",
        limits: tuple = (0, 1),
        agg: str = "mean",
        fillna=None,
        skipna=True,
    ):
        """Extract statistics on 2D variables.

        General purpose method for extracting statistics using typical statistic functions (e.g., 'mean', 'max'...).

        Percentiles can be obtained using the string format 'pX' (e.g., for 10th and 90th: stats = ['p10', 'p90']).

        This function will first call `get_sheet` to obtain the data for calculating statistics on. All arguments for `get_sheet` can be passed through to this method.

        Args:
            stats (list[str]): list of statistics, e.g., ['min', 'mean', 'p50', 'p95']
            variables ([list, str]): variables to do statistics on
            time ([list, slice], optional): time indexer (e.g., slice(50) for the first 50 timesteps, or slice('2012-01-01', '2012-02-01'))
            datum (['sigma', 'height', 'depth', 'elevation'], optional): depth-averaging datum. Defaults to 'sigma'. Choose from `height`, `depth`, `elevation` or `sigma`.
            limits (tuple, optional): depth-averaging limits. Defaults to (0,1).
            agg (['min', 'mean', 'max'], optional): depth-averaging aggregation function. Defaults to 'mean'. Choose from `min`, `mean`, or `max`.
            fillna (float, optional): Fill nan values (e.g., dried out cells) with this value before doing statistics. Defaults to None.
            skipna (bool, optional): Skip nan values. Defaults to True.

        Returns:
            TfvDomain: new TfvDomain object with the requested timeseries variables

        """

        # Convert variable if provided as a singular string
        if isinstance(variables, str):
            variables = [variables]

        # BUG: Variables is being modified in-place by self.get_sheet
        # I don't know why, this is a quick-fix
        reqvars = variables.copy()

        # Convert stat variable if provided as a singular string
        if isinstance(stats, str):
            stats = [stats]

        # Check if data needs extracting (i.e., have I been chained?)
        extract = False
        for v in variables:
            # Unfortunate check because of virtual variables ('V', 'VDir')
            if v in self.variables:
                if "NumCells3D" in self[v].dims:
                    extract = True
            else:
                extract = True

        # Extract data, if required (yes I've been chained)
        if extract:
            ds_2d = self.get_sheet(reqvars, time, datum, limits, agg)
        else:
            ds_2d = self

        if fillna:
            ds_2d = ds_2d.fillna(fillna)

        # Prep output dataset (blank, Time == 1)
        dso = _strip_dataset(self._obj.isel(Time=[0]))

        # Process the requested statistics
        funcs = []
        quantiles = []
        for x in stats:
            if re.match("p\d+", x):
                quantiles.append(float(re.findall(r"[-+]?(?:\d*\.*\d+)", x)[0]) / 100)
            else:
                funcs.append(x)

        # Run functions
        # Supress all-nan slice warning here - this is common for TFV outputs!
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", r"All-NaN (slice|axis) encountered")

            # Run built-in xarray functions
            for f in funcs:
                for v in variables:
                    vname = f"{v}_{f}"
                    try:
                        func = getattr(ds_2d[v], f)
                        dso[vname] = func(dim="Time", skipna=skipna).expand_dims("Time")
                        dso[vname].attrs = ds_2d[v].attrs
                    except AttributeError:
                        print(
                            f"{f} is not supported - see xarray documentation for available functions"
                        )

            # Run quantiles (all in one)
            if len(quantiles) > 0:
                for v in variables:
                    qarr = ds_2d[v].quantile(quantiles, skipna=skipna, dim="Time")
                    for ii, q in enumerate(quantiles):
                        vname = f"{v}_p{q*100:n}"
                        dso[vname] = qarr.isel(quantile=ii).expand_dims("Time")
                        dso[vname].attrs = ds_2d[v].attrs

        dso = dso.drop_vars("quantile", errors="ignore")
        dso.attrs = self._obj.attrs

        return dso.tfv

    def get_profile(
        self,
        point: Union[tuple, dict],
        time: slice = None,
        variables: list[str] = None,
    ):
        """Extract profile timeseries at a single location

        Simple method to extract a profile timeseries at a single location. This returns a native Xarray dataset, and does not have the methods available using the TfvTimeseries Xarray accessor.

        Args:
            point ([tuple, dict]): Single location to extract profile as a tuple (X,Y) or dictionary dict(loc_01=(X,Y)).
            time ([str, pd.Timestamp, int, slice], optional): time indexer for extraction. Defaults to the entire dataset.
            variables (list[str], optional): List of variables. Defaults to all variables.

        Returns:
            xr.Dataset: a native Xarray dataset containing the profile timeseries.

        """
        # Subset time
        xtr = self.xtr
        if time is not None:
            dsx = xtr._subset_dataset(time)
        else:
            dsx = xtr._subset_dataset(slice(None))

        # Convert point argument
        if isinstance(point, dict):
            assert (
                len(point) == 1
            ), "Can only extract ONE profile at a time, please reduce locations to a single {name: (x, y)} format"
            name = list(point.keys())[0]
            point = point[name]
        else:
            name = "loc01"

        # Convert variables if required
        if isinstance(variables, tuple):
            variables = list(variables)

        # Grab cell indicies
        i2 = xtr.get_cell_index(point[0], point[1])[0]
        i3 = np.where(self.xtr.idx2 == i2)[0]
        i4 = np.where(self.xtr.idx4 == i2)[0]
        if i2 == -999:
            print(f"WARNING: point '{name}' is outside of model domain")
        dsx = dsx.sel(NumCells2D=i2, NumCells3D=i3, NumLayerFaces3D=i4)

        # Refactor, this is yucky
        dsx = dsx.drop_vars(
            [
                "ResTime",
                "cell_Nvert",
                "NL",
                "cell_node",
                "cell_A",
                "node_X",
                "node_Y",
                "node_Zb",
                "idx2",
                "idx3",
                "node_NVC2",
                "node_cell2d_idx",
                "node_cell2d_weights",
            ],
            errors="ignore",
        )
        dsx = dsx.rename(
            dict(
                cell_X="X",
                cell_Y="Y",
                cell_Zb="Z",
                NumCells3D="NumLayers",
                NumLayerFaces3D="NumLayerFaces",
            )
        )

        dsx["Z"] = (
            ("Time", "NumLayers"),
            np.mean((dsx["layerface_Z"][:, :-1], dsx["layerface_Z"][:, 1:]), axis=0),
        )

        for var in dsx.data_vars.keys():
            if len(dsx[var].dims) == 1:
                dsx[var] = dsx[var].expand_dims(dim="N1", axis=1)

        # Add vel, if not already computed
        if "V" not in self.variables:
            dsx["V"] = np.hypot(dsx["V_x"], dsx["V_y"])
            dsx["V"].attrs = {"long_name": "current speed", "units": "m s^-1"}
        if "VDir" not in self.variables:
            dsx["VDir"] = (90 - np.arctan2(dsx["V_y"], dsx["V_x"]) * 180 / np.pi) % 360
            dsx["VDir"].attrs = {"long_name": "current direction", "units": "degN TO"}

        # Optional variable - drop unrequired variables, and re-order as requested
        if variables is not None:
            dsx = dsx[["X", "Y", "Z", "stat", "layerface_Z"] + variables]

        # Add N1 dim to match TUFLOW-FV Output
        dsx["X"] = dsx["X"].expand_dims(dim="N1")
        dsx["Y"] = dsx["Y"].expand_dims(dim="N1")

        dsx.attrs["Loc. name"] = name
        dsx.attrs["Loc. coords"] = point
        dsx.attrs[
            "Origin"
        ] = "Profile extracted from TUFLOWFV cell-centered output using `tfv` xarray accessor"
        dsx.attrs["Type"] = "Profile cell from TUFLOWFV output"

        return dsx

    def get_curtain(
        self,
        polyline: np.ndarray,
        variables: Union[str, list],
        time: Union[str, int, pd.Timestamp, slice] = None,
    ):
        """Extract a 2D curtain timeseries, returned as an Xarray dataset.

        Note:
            This is a work-in-progress method, and is not directly used in any other methods currently. If you wish to just view curtain plots, please use the `plot_curtain` method.

        Args:
            polyline (np.ndarray): a Nx2 array containing the X and Y coordinates to extract the curtain over.
            variables ([str, list]): single or list of variables to extract
            time ([str, pd.Timestamp, int, slice], optional): time indexer for extraction. Defaults to the entire dataset.

        Returns:
            xr.Dataset: Xarray dataset object
        """
        xtr = self.xtr
        # Use sliced xarray dataset
        dsx = xtr._subset_dataset(time)
        times = dsx["Time"]
        ts = [xtr._timestep_index(t) for t in times.values]
        nt = times.shape[0]

        if isinstance(variables, str):
            variables = [variables]

        if ~isinstance(polyline, np.ndarray):
            polyline == np.asarray(polyline)
        assert (
            polyline.shape[1] == 2
        ), "Polyline should be an Nx2 numpy array, or a list of coordinates [[x1,y1], [x2,y2], ..., [xn, yn]]"

        line_index, line_cell_3D = xtr.get_curtain_cell_index(polyline)
        nx, nz, tri = xtr.get_curtain_cell_geo(ts[0], polyline)
        x_data = xtr.get_intersection_data(polyline)
        ns = tri.shape[0]  # Num cell
        # print(x_data[2])

        cells_2d = x_data[2][x_data[2] >= 0]
        cell_X = self.geo["cell_X"][cells_2d]
        cell_Y = self.geo["cell_Y"][cells_2d]
        cell_Zb = self.geo["cell_Zb"][cells_2d]

        if self._obj.attrs.get("Spherical", "true") == "true":
            unit = "decimal degrees"
        else:
            unit = "m"

        dstx = xr.Dataset(
            coords=dict(
                Time=(("Time",), times.values),
            ),
            data_vars=dict(
                node_X=(("NumVert1D"), x_data[0]),
                node_Y=(("NumVert1D"), x_data[1]),
                cell_X=(("NumCells1D"), cell_X),
                cell_Y=(("NumCells1D"), cell_Y),
                cell_Zb=(("NumCells1D"), cell_Zb),
                line_cells_2D=(("NumCells1D"), cells_2d),
                line_cells_3D=(("NumVert2D"), line_cell_3D),
                line_index=(("NumVert2D"), line_index),
                Chainage=(("NumVert2D", "MaxNumCellVert"), nx[tri]),
                Z=(("Time", "NumVert2D", "MaxNumCellVert"), np.zeros((nt, ns, 4))),
            ),
        )
        dstx["Chainage"].attrs = {"long_name": "chainage", "units": "m"}
        dstx["cell_X"].attrs = {
            "long_name": "Cell Centroid X-Coordinate",
            "units": unit,
        }
        dstx["cell_Y"].attrs = {
            "long_name": "Cell-Centroid Y-Coordinate",
            "units": unit,
        }
        dstx["cell_Zb"].attrs = {"long_name": "Cell Bed Elevation", "units": "m"}
        dstx["node_X"].attrs = {"long_name": "Node X-Coordinate", "units": unit}
        dstx["node_Y"].attrs = {"long_name": "Node Y-Coordinate", "units": unit}
        dstx["line_cells_2D"].attrs = {
            "long_name": "Cell-Centroid Y-Coordinate",
            "units": unit,
        }
        dstx["line_cells_3D"].attrs = {
            "long_name": "Cell-Centroid Y-Coordinate",
            "units": unit,
        }
        dstx["line_index"].attrs = {
            "long_name": "Cell-Centroid Y-Coordinate",
            "units": unit,
        }
        dstx["Z"].attrs = {"long_name": "elevation in water column", "units": "m"}

        # Init variables
        for v in variables:
            dstx[v] = (("Time", "NumVert2D"), np.zeros((nt, ns)))
            if v in self.variables:
                assert "NumCells3D" in dsx[v].dims, f'"{v}" is not a 3D variable'
                dstx[v].attrs = dsx[v].attrs
            elif v == "V":
                dstx[v].attrs = {"long_name": "current speed", "units": "m s^-1"}
            elif v == "VDir":
                dstx[v].attrs = {"long_name": "current direction", "units": "degN TO"}

        # Add V/VDir methods
        requested_vars = variables.copy()
        if "V" in variables:
            variables.extend(["V_x", "V_y"])
            variables.pop(variables.index("V"))
        if "VDir" in variables:
            variables.extend(["V_x", "V_y"])
            variables.pop(variables.index("VDir"))
        variables = np.unique(variables).tolist()

        # Init output array
        nv = len(variables)
        array = np.ma.zeros((nv, nt, ns))

        c = 0
        for t in tqdm(
            ts,
            desc="...extracting sheet data",
            colour="green",
            delay=1,
        ):
            # Fetch elevation
            nx, nz, tri = xtr.get_curtain_cell_geo(ts[0], polyline)
            dstx["Z"][c, :] = nz[tri]

            array[:, c, :] = xtr.get_curtain_cell(variables, t, polyline)
            c += 1
        for v in requested_vars:
            if v == "V":
                vxi = variables.index("V_x")
                vyi = variables.index("V_y")
                arr = np.hypot(array[vxi], array[vyi])
                dstx[v].values = arr

            elif v == "VDir":
                vxi = variables.index("V_x")
                vyi = variables.index("V_y")
                arr = (90 - np.arctan2(array[vyi], array[vxi]) * 180 / np.pi) % 360
                dstx[v].values = arr

            else:
                dstx[v].values = array[variables.index(v)]

        dstx.attrs = {
            "Origin": "Curtain extracted from TUFLOWFV Output",
            "Type": "Curtain TUFLOWFV output",
        }

        return dstx

    def get_timeseries(
        self,
        variables: Union[str, list],
        points: Union[tuple, dict],
        time: Union[str, int, pd.Timestamp, slice] = None,
        datum: Literal["sigma", "height", "depth", "elevation"] = "sigma",
        limits: tuple = (0, 1),
        agg: Literal["min", "mean", "max"] = "mean",
    ):
        """Get timeseries at point location(s) in the model

        Args:
            variables ([str, list]): variables to extract.
            points ([tuple, dict]): locations to extract (e.g., a tuple (x,y) or a dict {'name': (x,y), ...})
            time ([str, pd.Timestamp, int, slice], optional): time indexer for extraction. Defaults to the entire dataset.
            datum (['sigma', 'height', 'depth', 'elevation'], optional): depth-averaging datum. Defaults to 'sigma'. Choose from `height`, `depth`, `elevation` or `sigma`.
            limits (tuple, optional): depth-averaging limits. Defaults to (0,1).
            agg (['min', 'mean', 'max'], optional): depth-averaging aggregation function. Defaults to 'mean'. Choose from `min`, `mean`, or `max`.

        Returns:
            xr.Dataset: A timeseries dataset in xarray format
        """
        # Use sliced xarray dataset
        dsx = self.xtr._subset_dataset(time)

        if isinstance(variables, str):
            variables = [variables]
        # Get 2D cell index
        if isinstance(points, dict):
            index = []
            for pt in points.values():
                index.append(int(self.xtr.get_cell_index(pt[0], pt[1])))
            if len(points) == 1:
                pt = list(points.values())[0]
                index = self.xtr.get_cell_index(pt[0], pt[1])
                points = pt
        else:
            index = self.xtr.get_cell_index(points[0], points[1])
            assert index >= 0, "Point is not inside model domain"

        dst = xr.Dataset(coords=dict(Time=dsx.Time), attrs=dsx.attrs)
        dst.attrs[
            "Origin"
        ] = "Timeseries extracted from TUFLOWFV cell-centered output using `tfv` python tools"
        dst.attrs["Type"] = "Timeseries cell from TUFLOWFV Output"
        dst.attrs["Datum"] = str(datum)
        dst.attrs["Limits"] = str(limits)
        dst.attrs["Agg Fn"] = agg

        # Add coordinate locations
        if len(index) > 1:
            dst = dst.assign_coords(
                dict(
                    Location=(("Location"), [x for x in points.keys()]),
                    x=(("Location"), [x[0] for x in points.values()]),
                    y=(("Location"), [x[1] for x in points.values()]),
                    z=(("Location"), self.geo["cell_Zb"][index]),
                )
            )
        else:
            dst = dst.assign_coords(
                dict(
                    Location=(("Location"), [0]),
                    x=(("Location",), [points[0]]),
                    y=(("Location",), [points[1]]),
                )
            )

        # Add output variables to dataset
        for v in variables:
            dst[v] = (("Time", "Location"), np.zeros((dst.dims["Time"], len(index))))

            if v not in ["V", "VDir"]:
                dst[v].attrs = dsx[v].attrs
            elif v == "V":
                dst[v].attrs = {"long_name": "current speed", "units": "m s^-1"}
            elif v == "VDir":
                dst[v].attrs = {"long_name": "current direction", "units": "degN TO"}

        ii = 0
        for t in tqdm(
            dsx["Time"].values, desc="Extracting timeseries, please wait", colour="blue"
        ):
            for v in variables:
                val = self.xtr.get_sheet_cell(v, t, datum, limits, agg)[index]
                dst[v][ii] = val
            ii += 1

        if len(index) == 1:
            dst = dst.isel(Location=0)

        return dst

    def plot(
        self,
        variable: str = None,
        time: Union[str, pd.Timestamp, int] = 0,
        colorbar=True,
        shading="patch",
        ax: plt.Axes = None,
        **kwargs,
    ):
        """General 2D sheet plot method

        By default, this function will plot the first variable and timestep.

        This function is a general wrapper around the tfv.visual "Sheet" methods. Kwargs will be supplied through to the underlying matplotlib function.

        Depth-averaging logic can be handled by passing through the optional keywords `datum, limits, agg` as required.

        Several magic "variables" are available via this method including:
            - Plot bed elevation using variable 'Zb'
            - Plot current speed or direction using either 'V' or 'VDir' respectively

        Args:
            variable (str, optional): variable to plot. Defaults to alphabetical first.
            time ([str, pd.Timestamp, int], optional): Timestep to plot, provided as str (e.g., '2010-01-01'), int, Timestamp object, etc. Defaults to 0.
            colorbar (bool, optional): Show colorbar (best to turn off if making subplots). Defaults to True.
            shading (str, optional): Figure shading type, which changes the underlying matplotlib function.
                Options include:
                    "patch" that calls tfv.visual.SheetPatch and Matplotlib's `PolyCollection`
                    "interp" that calls tfv.visual.SheetPatch and Matplotlib's `TriMesh`
                    "contour" that calls tfv.visual.SheetContour and Matplotlib's `TriContourSet`
                Defaults to "patch".
            ax (plt.Axes, optional): Matplotlib axis to draw on. Default will create a new figure and axis.
            kwargs (dict, optional): Keyword arguments. These are passed to the underlying matplotlib method if not a tfv specific method (e.g., depth averaging kwargs).
                Common examples include: cmap, clim, levels, norm, datum, limits, agg

        Returns:
            Sheet: tfv.visual.Sheet<X> Object, depending on the shading option.
        """
        variable, label = self._getvar_(variable)

        # Special cases (data without a time dimension)
        if variable == "Zb":
            if shading == "patch":
                variable = self._obj["cell_Zb"].values
            else:
                variable = self._obj["node_Zb"].values
            static = True
        else:
            static = False

        fig, ax, zoom = _prep_axis(ax, kwargs)

        depth_kwargs = dict(
            datum=kwargs.pop("datum", "sigma"),
            limits=kwargs.pop("limits", (0, 1)),
            agg=kwargs.pop("agg", "mean"),
        )

        if shading == "patch":
            edgecolor = kwargs.pop("edgecolor", "face")
            sheet = SheetPatch(
                ax,
                self.xtr,
                variable,
                edgecolor=edgecolor,
                zoom=zoom,
                **depth_kwargs,
                **kwargs,
            )
            mappable = sheet.patch

        elif shading == "interp":
            sheet = SheetPatch(
                ax,
                self.xtr,
                variable,
                shading=shading,
                zoom=zoom,
                **depth_kwargs,
                **kwargs,
            )
            mappable = sheet.patch

        elif shading == "contour":
            sheet = SheetContour(
                ax, self.xtr, variable, zoom=zoom, **depth_kwargs, **kwargs
            )
            mappable = sheet.cont

        # Update to current time
        if static == False:
            sheet.set_time_current(time)

            date_title = sheet.get_time_current().strftime("%Y-%m-%d %H:%M")
            ax.set_title(date_title)

        if colorbar:
            fig.colorbar(mappable, ax=ax, label=label)

        # Set tick format to float
        ax.yaxis.set_major_locator(plt.MaxNLocator(5))
        ax.yaxis.set_major_formatter(FormatStrFormatter("%.3f"))
        ax.xaxis.set_major_locator(plt.MaxNLocator(5))
        ax.xaxis.set_major_formatter(FormatStrFormatter("%.3f"))

        if ("ipympl" in matplotlib.get_backend()) & (shading != "contour"):
            z = mappable.get_array()
            xf = self.geo["cell_X"].flatten()
            yf = self.geo["cell_Y"].flatten()
            zf = z.flatten()

            def fmt_coord(x, y):
                dist = np.linalg.norm(np.vstack([xf - x, yf - y]), axis=0)
                idx = np.argmin(dist)
                z = zf[idx]
                return f"x={x:.5f}  y={y:.5f}  var={z:.5f}"

            plt.gca().format_coord = fmt_coord

        return sheet

    def plot_vector(
        self,
        variables: Union[list, tuple] = None,
        time: Union[str, pd.Timestamp, int] = 0,
        plot_type: Literal["quiver"] = "quiver",  # ToDO: Streamplot ya dog
        ax: plt.Axes = None,
        **kwargs,
    ):
        """General 2D vector plot method

        Generic method to call tfv.visual's `SheetVector`. This can be used to overlay vector plots on other figures.
        By default, this method will plot vectors using V_x and V_y to give velocity vectors, however a list/tuple of an x/y component can be provided.

        kwargs passed to this function are supplied to the underlying matplotlib function (e.g., Quiver or Streamplot).
        The default vector scaling for velocity is usually ok; however in some cases you may need to scale the vectors (e.g., scale=100).

        Vectors can also be set to a uniform length using `normalised=True`.

        Depth-averaging logic can be handled by passing through the optional keywords `datum, limits, agg` as required.

        Args:
            variables ([list, tuple], optional): Variables to create vector (X-comp, Y-comp). Defaults to None.
            time ([str, pd.Timestamp, int], optional): Timestep to plot, provided as str (e.g., '2010-01-01'), int, Timestamp object, etc. Defaults to 0.
            plot_type (Literal['quiver'], optional): _description_. Defaults to "quiver".
            ax (plt.Axes, optional): Matplotlib axis to draw on. Default will create a new figure and axis.
            kwargs (dict, optional): Keyword arguments. These are passed to the underlying matplotlib method if not a tfv specific method (e.g., depth averaging kwargs).
                Common examples include: scale, normalised, datum, limits, agg.

        Returns:
            SheetVector: `tfv.visual.SheetVector` object.
        """

        if (variables is None) | (variables == "V"):
            variables = ["V_x", "V_y"]

        fig, ax, zoom = _prep_axis(ax, kwargs)

        depth_kwargs = dict(
            datum=kwargs.pop("datum", "sigma"),
            limits=kwargs.pop("limits", (0, 1)),
            agg=kwargs.pop("agg", "mean"),
        )

        if plot_type == "quiver":
            vectors = SheetVector(
                ax, self.xtr, variables, zorder=99, zoom=zoom, **depth_kwargs, **kwargs
            )
            # BUG: This is required to make SheetVector respect axis xlims. Fix this later
            vectors.__static_update__()

        # Update to current time
        vectors.set_time_current(time)

        date_title = vectors.get_time_current().strftime("%Y-%m-%d %H:%M")
        ax.set_title(date_title)

        return vectors

    def plot_curtain(
        self,
        polyline: np.ndarray,
        variable: str = None,
        time: Union[str, pd.Timestamp, int] = 0,
        ax: plt.Axes = None,
        colorbar=True,
        **kwargs,
    ):
        """General 2D curtain  plot method

        By default, this function will plot the first variable and timestep. A polyline is always required.

        This function is a general wrapper around the tfv.visual.CurtainPatch method. kwargs will be supplied through to the underlying matplotlib function.

        A polyline (Nx2 `numpy.ndarray`) needs to be supplied containing the X and Y coordinates in columns 0 and 1 respectively.

        Args:
            polyline (np.ndarray): a Nx2 array containing the X and Y coordinates to extract the curtain over.
            variable (str, optional): variable to plot. Defaults to 'V' if available, otherwise the first variable alphabetically.
            time ([str, pd.Timestamp, int], optional): Timestep to plot, provided as str (e.g., '2010-01-01'), int, Timestamp object, etc. Defaults to 0.
            ax (plt.Axes, optional): Matplotlib axis to draw on. Default will create a new figure and axis.
            colorbar (bool, optional): Show colorbar (best to turn off if making subplots). Defaults to True.

        Returns:
            CurtainPatch: `tfv.visual.CurtainPatch` object
        """
        if variable is None:
            if ("V_x" in self.variables) & ("V_y" in self.variables):
                variable = "V"
            elif self.variables[0] != "H":
                variable = self.variables[0]
            else:
                raise ValueError(
                    "Unable to detect what variable to plot - please specify a variable"
                )

        if variable == "V":
            attrs = {"long_name": "current speed", "units": "m s^-1"}
        elif variable == "VDir":
            attrs = {"long_name": "current direction", "units": "degN TO"}
        else:
            attrs = self[variable].attrs

        if ("long_name" in attrs) & ("units" in attrs):
            label = f"{attrs['long_name']} ({attrs['units']})"

        fig, ax, zoom = _prep_axis(ax, kwargs, equal=False)

        curtain = CurtainPatch(ax, self.xtr, variable, polyline, zoom=zoom, **kwargs)

        # Update to current time
        curtain.set_time_current(time)

        date_title = curtain.get_time_current().strftime("%Y-%m-%d %H:%M")
        ax.set_title(date_title)
        ax.set_ylabel("Elevation (m)")
        ax.set_xlabel("Chainage (m)")

        if colorbar:
            fig.colorbar(curtain.patch, ax=ax, label=label)

        return curtain

    def plot_curtain_vector(
        self,
        polyline: np.ndarray,
        time: Union[str, pd.Timestamp, int] = 0,
        variables: Union[list, tuple] = None,
        ax: plt.Axes = None,
        tangential=True,
        **kwargs,
    ):
        """General 2D curtain vector plot method

        This method is designed to draw velocity vectors over a curtain. If `W` (vertical velocity) is available in the TUFLOW FV dataset, these vectors will be scaled to show vertical motion as well.
        To avoid vertical velocity scaling of the vectors, specify `tangential=False` as well.

        Note:
            This method accepts a `variables` kwarg so that custom variables can be supplied, which will be scaled with vertical velocity `W` if present.
            To avoid this behaviour, also specify `tangential=False`.

        This function is a general wrapper around the tfv.visual.CurtainVector method. kwargs will be supplied through to the underlying matplotlib function.

        A polyline (Nx2 `numpy.ndarray`) needs to be supplied containing the X and Y coordinates in columns 0 and 1 respectively.

        Args:
            polyline (np.ndarray): a Nx2 array containing the X and Y coordinates to extract the curtain over.
            time ([str, pd.Timestamp, int], optional): Timestep to plot, provided as str (e.g., '2010-01-01'), int, Timestamp object, etc. Defaults to 0.
            variables ([list, tuple], optional): Variables to create vector (X-comp, Y-comp). Defaults to None.
            ax (plt.Axes, optional): Matplotlib axis to draw on. Default will create a new figure and axis.
            tangential (bool, optional): Flag to scale vectors vertically using `W`, if present. Defaults to True.

        Returns:
            CurtainVector: `tfv.visual.CurtainVector` object.
        """
        if variables is None:
            variables = ["V_x", "V_y"]

        fig, ax, zoom = _prep_axis(ax, kwargs, equal=False)

        curtain = CurtainVector(
            ax,
            self.xtr,
            variables,
            polyline,
            zoom=zoom,
            tangential=tangential,
            **kwargs,
        )

        # Update to current time
        curtain.set_time_current(time)

        date_title = curtain.get_time_current().strftime("%Y-%m-%d %H:%M")
        ax.set_title(date_title)
        ax.set_ylabel("Elevation (m)")
        ax.set_xlabel("Chainage (m)")

        return curtain

    def plot_interactive(
        self,
        variable: str,
        ax: plt.Axes = None,
        vectors: Union[bool, list, tuple] = False,
        vector_kwargs={},
        widget=None,
        **kwargs,
    ):
        """2D interactive matplotlib plotting method for fast result viewing.

        This function wraps the `.plot` and `.plot_vector` methods, with kwargs passed directly through to these methods.
        (see `TfvDomain.plot` e.g., ds.tfv.plot and TfvDomain.plot_vector e.g., ds.tfv.plot_vector).

        This function requires matplotlib to be using an ipympl backend, typically in a Jupyter lab/notebook environment.
        Please first run `%matplotlib widget` before using this function.

        Args:
            variable (str): Variable to plot
            ax (plt.Axes, optional): Matplotlib axis to draw on. Default will create a new figure and axis.
            vectors ([bool, list, tuple], optional): Flag to draw vectors. If a boolean (True) is supplied, the vectors will be velocity (V_x, V_y).
                A list/tuple of variable names can be supplied as required (E.g., W10_x, W10_y). Defaults to False.
            vector_kwargs (dict, optional): a dictionary of kwargs that is passed directly through to the `TfvDomain.plot_vector` object.
            widget (tuple, optional): A pre-initalised ipympl widget box, generated using `TfvDomain.prep_interactive_slider`.
                This can be used to control multiple subplots using the single widget controller. Defaults to None.
            kwargs (dict, optional): Keyword arguments passed directly to `TfvDomain.plot`
        """
        # Setup simple interactive plot
        time_vec = pd.to_datetime(self["Time"].values)
        fmt = "%Y-%m-%d %H:%M"

        # Prepare a widget instance
        if widget is None:
            grid = self.prep_interactive_slider()
        else:
            grid = widget
        slider = grid[1, 0].children[0]

        prev_arrow = grid[0, 0].children[1]
        next_arrow = grid[0, 0].children[2]
        date_picker = grid[2, 0].children[0]
        date_submit = grid[2, 0].children[1]

        fig, ax, zoom = _prep_axis(ax, kwargs)

        shading = kwargs.pop("shading", "patch")
        depth_kwargs = dict(
            datum=kwargs.pop("datum", "sigma"),
            limits=kwargs.pop("limits", (0, 1)),
            agg=kwargs.pop("agg", "mean"),
        )

        sheet = self.plot(
            variable, 1, ax=ax, zoom=zoom, shading=shading, **depth_kwargs, **kwargs
        )
        if vectors is not None:
            if vectors == True:
                assert all(
                    ("V_x" in self.variables, "V_y" in self.variables)
                ), "Vector plot expected `V_x` & `V_y` in dataset, however these are not present. Please check, or supply a list [<Var_x>, <Var_y>]"
                vec = self.plot_vector(
                    ["V_x", "V_y"], 1, ax=ax, zoom=zoom, **depth_kwargs, **vector_kwargs
                )
            elif any([isinstance(vectors, x) for x in [list, tuple]]):
                vec = self.plot_vector(
                    vectors, 1, ax=ax, zoom=zoom, **depth_kwargs, **vector_kwargs
                )
                vectors = True
        ax.set_title("")

        xf = self.geo["cell_X"].flatten()
        yf = self.geo["cell_Y"].flatten()

        def update_time(change):
            date_picker.value = time_vec[change.new]

            # ax.set_title(f"{time_vec[change.new].strftime(fmt)}")

            sheet.set_time_current(change.new)
            if vectors == True:
                vec.set_time_current(change.new)

            # Update z-coordinate boilerplate
            if shading != "contour":
                zf = sheet.patch.get_array().flatten()

                def fmt_coord(x, y):
                    dist = np.linalg.norm(np.vstack([xf - x, yf - y]), axis=0)
                    idx = np.argmin(dist)
                    z = zf[idx]
                    return f"x={x:.5f}  y={y:.5f}  {variable}={z:.5f}"

                ax.format_coord = fmt_coord

            plt.draw()

        def submit_date(btn):
            c = np.argmin(np.abs(time_vec - date_picker.value))
            slider.value = c

        def next_ts(btn):
            slider.value += 1

        def prev_ts(btn):
            slider.value -= 1

        slider.observe(update_time, names="value")
        prev_arrow.on_click(prev_ts)
        next_arrow.on_click(next_ts)
        date_submit.on_click(submit_date)

        slider.value = 1
        if widget is None:
            from IPython.display import display

            return display(grid)

    def plot_curtain_interactive(
        self,
        polyline: np.ndarray,
        variable: str,
        ax: plt.Axes = None,
        vectors: bool = False,
        vector_kwargs={},
        widget=None,
        **kwargs,
    ):
        """2D interactive matplotlib method for fast curtain viewing.

        This function wraps the `.plot_curtain` and `.plot_curtain_vector` methods, with kwargs passed directly through to these methods.
        (see `TfvDomain.plot_curtain` e.g., ds.tfv.plot and TfvDomain.plot_curtain_vector e.g., ds.tfv.plot_curtain_vector).
        
        A polyline (Nx2 `numpy.ndarray`) needs to be supplied containing the X and Y coordinates in columns 0 and 1 respectively.

        This function requires matplotlib to be using an ipympl backend, typically in a Jupyter lab/notebook environment.
        Please first run `%matplotlib widget` before using this function.
        
        Args:
            polyline (np.ndarray): a Nx2 array containing the X and Y coordinates to extract the curtain over.
            variable (str): Variable to plot
            ax (plt.Axes, optional): Matplotlib axis to draw on. Default will create a new figure and axis.
            vectors ([bool, list, tuple], optional): Flag to draw vectors. If `True`, the vectors will represent projected velocity  (V_x, V_y projected to the curtain plane). Vertical velocity, W, will be included automatically if available in the model.
            vector_kwargs (dict, optional): a dictionary of kwargs that is passed directly through to the `TfvDomain.plot_curtain_vector` object.
            widget (tuple, optional): A pre-initalised ipympl widget box, generated using `TfvDomain.prep_interactive_slider`.
                This can be used to control multiple subplots using the single widget controller. Defaults to None.
            kwargs (dict, optional): Keyword arguments passed directly to `TfvDomain.plot_curtain`
        """
        # Setup simple interactive plot
        time_vec = pd.to_datetime(self["Time"].values)
        fmt = "%Y-%m-%d %H:%M"

        # Prepare a widget instance
        if widget is None:
            grid = self.prep_interactive_slider()
        else:
            grid = widget
        slider = grid[1, 0].children[0]

        prev_arrow = grid[0, 0].children[1]
        next_arrow = grid[0, 0].children[2]
        date_picker = grid[2, 0].children[0]
        date_submit = grid[2, 0].children[1]

        fig, ax, zoom = _prep_axis(ax, kwargs, equal=False)

        curtain = self.plot_curtain(polyline, variable, 1, ax=ax, **kwargs)
        if vectors == True:
            assert all(
                ("V_x" in self.variables, "V_y" in self.variables)
            ), "Curtain vectors requires `V_x` & `V_y` in dataset!"
            vec = self.plot_curtain_vector(polyline, time=1, ax=ax, **vector_kwargs)
        ax.set_title("")

        # xf = self.geo["cell_X"].flatten()
        # yf = self.geo["cell_Y"].flatten()

        def update_time(change):
            date_picker.value = time_vec[change.new]

            # ax.set_title(f"{time_vec[change.new].strftime(fmt)}")

            curtain.set_time_current(change.new)
            if vectors == True:
                vec.set_time_current(change.new)

            # # Update z-coordinate boilerplate
            # if shading != "contour":
            #     zf = sheet.patch.get_array().flatten()

            #     def fmt_coord(x, y):
            #         dist = np.linalg.norm(np.vstack([xf - x, yf - y]), axis=0)
            #         idx = np.argmin(dist)
            #         z = zf[idx]
            #         return f"x={x:.5f}  y={y:.5f}  {variable}={z:.5f}"

            #     ax.format_coord = fmt_coord

            plt.draw()

        def submit_date(btn):
            c = np.argmin(np.abs(time_vec - date_picker.value))
            slider.value = c

        def next_ts(btn):
            slider.value += 1

        def prev_ts(btn):
            slider.value -= 1

        slider.observe(update_time, names="value")
        prev_arrow.on_click(prev_ts)
        next_arrow.on_click(next_ts)
        date_submit.on_click(submit_date)

        slider.value = 1
        if widget is None:
            from IPython.display import display

            return display(grid)

    def prep_interactive_slider(self):
        """Prepare an interactive ipympl widget box for controlling `TfvDomain.plot_interactive`

        This method pre-initialises a widget box that can be used to control the interactive jupyter plotting method, and is called by default from the `plot_interactive` method.

        To control multiple interactive plots using the one widget box, first call this method, e.g. `widget=ds.tfv.prep_interactive_slider()`, and then pass this widget through to the `plot_interactive` method.
        """
        try:
            import ipympl
            from ipywidgets import widgets, GridspecLayout, Layout
            import matplotlib
        except ImportError:
            print(
                "The interactive plot requries that `ipympl` and `ipywidgets` be installed"
            )

        assert (
            "ipympl" in matplotlib.get_backend()
        ), "Please run `%matplotlib widget` to enable interactive plotting in your Jupyter session!"

        nt = self.dims["Time"]
        time_vec = pd.to_datetime(self["Time"].values)

        # Slider
        slider = widgets.IntSlider(
            value=0, min=0, max=nt, step=1, layout=Layout(width="40%", height="85%")
        )

        play = widgets.Play(
            value=0,
            min=0,
            max=nt,
            step=1,
            interval=500,
            description="Play",
            disabled=False,
            layout=Layout(width="10%", height="85%"),
        )


        widgets.jslink((play, "value"), (slider, "value"))
        play_slider = widgets.HBox([slider])

        # Date picker box
        date_picker = widgets.NaiveDatetimePicker(
            value=time_vec[0],
            min=time_vec[0],
            max=time_vec[-1],
            description="",
            style=dict(width="initial"),
            layout=Layout(width="260px", height="10px"),
        )

        date_submit = widgets.Button(
            description="Update",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Update",
            icon="arrow-turn-down-left",  # (FontAwesome names without the `fa-` prefix)
            style=dict(width="initial", height="20px"),
        )

        next_arrow = widgets.Button(
            description="",
            disabled=False,
            button_style="success",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="",
            icon="arrow-right",  # (FontAwesome names without the `fa-` prefix)
            layout=Layout(width="2.5%", height="85%"),
        )

        prev_arrow = widgets.Button(
            description="",
            disabled=False,
            button_style="danger",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="",
            icon="arrow-left",  # (FontAwesome names without the `fa-` prefix)
            layout=Layout(width="2.5%", height="85%"),
        )

        # date_box = widgets.HBox(
        #     [play, prev_arrow, next_arrow, date_picker, date_submit]
        # )

        # Prep gridbox
        grid = GridspecLayout(3, 1)
        grid[0, 0] = widgets.HBox([play, prev_arrow, next_arrow])
        grid[1, 0] = play_slider
        grid[2, 0] = widgets.HBox([date_picker, date_submit])

        return grid


class TfvTimeseries(TfvBase):
    """Xarray accessor object for working with TUFLOW FV timeseries profile netcdf files.

    Extends the functionality of native xarray to assist with navigating and extracting data from the grouped netcdf profile timeseries files.

    To use, call the `.tfv` method on an xarray dataset based on a TUFLOW FV timeseries file.
    """

    def __init__(self, xarray_obj):
        TfvBase.__init__(self, xarray_obj)
        self.ts = None
        self.__load_tfv_timeseries()

    def __repr__(self):
        if is_notebook is False:
            fmt = "%Y-%m-%d %H:%M:%S"
            # return self._obj.__repr__()
            print(" --- TUFLOW-FV Profile Output Dataset ---")
            print("")
            print(f"Num Timesteps: {self.nt}")
            print(f"Time start: {self.time_start.strftime(fmt)}")
            print(f"Time end: {self.time_end.strftime(fmt)}")
            if isinstance(self.time_step, int):
                print(f"Time step: {self.time_step} seconds")
            else:
                print(f"Time step: {self.time_step.mean().seconds} seconds")
            print("")
            print(f"Num Locations: {len(self.locations)}")
            print(f'Locations: {", ".join(self.locations.keys())}')
            print("")
            print(f'Variables in file: {", ".join(self.variables.keys())}')
            print("")
            print("Available methods: ")
            print("\t .get_timeseries(locations, variables, time, datum, limits, agg)")
            print("\t .get_location(location)")
            print("\t .plot(variable, location, time=None, ax=None)")
            return ""
        else:
            return "TUFLOW FV profile timeseries accessor object"

    def _repr_html_(self):
        """For now, this repr output is identical"""
        fmt = "%Y-%m-%d %H:%M:%S"
        # return self._obj.__repr__()
        print(" --- TUFLOW-FV Profile Output Dataset ---")
        print("")
        print(f"Num Timesteps: {self.nt}")
        print(f"Time start: {self.time_start.strftime(fmt)}")
        print(f"Time end: {self.time_end.strftime(fmt)}")
        if isinstance(self.time_step, int):
            print(f"Time step: {self.time_step} seconds")
        else:
            print(f"Time step: {self.time_step.mean().seconds} seconds")
        print("")
        print(f"Num Locations: {len(self.locations)}")
        print(f'Locations: {", ".join(self.locations.keys())}')
        print("")
        print(f'Variables in file: {", ".join(self.variables.keys())}')
        print("")
        print("Available methods: ")
        print("\t .get_timeseries(locations, variables, time, datum, limits, agg)")
        print("\t .get_location(location)")
        print("\t .plot(variable, location, time=None, ax=None)")
        return "TUFLOW FV profile timeseries accessor object"

    def __load_tfv_timeseries(self):
        if self.ts is None:
            try:
                self.ts = FvTimeSeries(self._obj)
                self.time = self.ts.time_vector
                self.locations = self.ts.locations
                self.nl = len(self.locations)
                self.nt = len(self.time)
                self.time_start = self.time[0]
                self.time_end = self.time[-1]
                self.time_duration = self.time[-1] - self.time[0]

                ts = pd.to_timedelta(
                    np.diff(self.time).astype(float) / 10**9, unit="S"
                )
                if len(set(ts)) == 1:
                    self.time_step = ts[0].seconds
                elif len(set(ts)) > 1:
                    self.time_step = ts
                else:
                    assert False, "No time data has been detected"

                # Override the original dataset with a more informative one.
                dsx = self._obj
                dsx = dsx.assign_coords(
                    Locations=(("Locations",), list(self.locations.keys()))
                )
                dsx = dsx.assign_coords(Time=(("Time",), self.time))
                coords = np.asarray(list(self.locations.values()))
                dsx = dsx.assign(X=(("Locations",), coords[:, 0]))
                dsx = dsx.assign(Y=(("Locations",), coords[:, 1]))
                dsx = dsx.assign(Variables=(("",), self.ts.variables))
                self._obj = dsx

            except TypeError:
                print(
                    "Data does not appear to be a valid TUFLOW-FV profile timeseries netcdf file"
                )

    @property
    def variables(self):
        grp = next(iter(self.locations.items()))[0]
        return {
            k: v.attrs
            for k, v in self.ds[grp].data_vars.items()
            if k not in ["X", "Y", "ResTime", "stat", "layerface_Z"]
            if "Time" not in v.attrs  # Check usual suspects, and Time
        }

    @property
    def ds(self):
        return self.ts.ds

    def get_location(self, location: str) -> xr.Dataset:
        """Get individual profile dataset

        Returns a native Xarray dataset for a single profile location.
        This can be alternatively accessed via a dictionary `.tfv.ds[location]`

        Args:
            locations (str): Location name to extract. (see `.locations` method for the list of stored locations)

        Returns:
            xr.Dataset: Individual profile dataset
        """

        assert (
            location in self.locations.keys()
        ), "Requested location not present in dataset"

        return self.ds[location]

    def get_timeseries(
        self,
        locations: Union[str, list] = None,
        variables: Union[str, list] = None,
        time: Union[str, int, pd.Timestamp, slice] = None,
        datum: Literal["sigma", "height", "depth", "elevation"] = "sigma",
        limits: tuple = (0, 1),
        agg: Literal["min", "mean", "max"] = "mean",
    ):
        """Extract 1D timeseries at location(s).

        Method to extract 1D timeseries at one, or several locations from the dataset.
        This method will handle dimension reduction, by default using depth-averaging.

        Note:
            This method can be slow for many locations and long-timeseries, as each profile location is effectively a standalone dataset.

        Args:
            locations (Union[str, list]): Location names to extract (see `.locations` method for the list of stored locations)
            variables ([str, list]): variables to extract. Defaults to all. ("V" or "VDir" may be requested if "V_x" and "V_y" are present).
            time ([str, pd.Timestamp, int, slice], optional): time indexer for extraction. Defaults to the entire dataset.
            datum (['sigma', 'height', 'depth', 'elevation'], optional): depth-averaging datum. Defaults to 'sigma'. Choose from `height`, `depth`, `elevation` or `sigma`.
            limits (tuple, optional): depth-averaging limits. Defaults to (0,1).
            agg (['min', 'mean', 'max'], optional): depth-averaging aggregation function. Defaults to 'mean'. Choose from `min`, `mean`, or `max`.

        Returns:
            xr.Dataset: A timeseries dataset in xarray format
        """

        if isinstance(variables, str):
            variables = [variables]

        # Get 2D cell index
        if isinstance(locations, str):
            locations = [locations]
        elif locations is None:
            locations = list(self.locations.keys())

        ds_set = []
        for l in locations:
            ds_set.append(
                self.ts.get_timeseries(
                    l,
                    variables=variables,
                    time=time,
                    datum=datum,
                    limits=limits,
                    agg=agg,
                )
            )
        ds = xr.concat(ds_set, dim="Location", combine_attrs="drop_conflicts")

        # Add coordinate locations
        ds = ds.assign_coords(
            dict(
                Location=(("Location"), locations),
                X=(("Location"), [self.locations[x][0] for x in locations]),
                Y=(("Location"), [self.locations[x][1] for x in locations]),
            )
        )
        return ds

    def plot(
        self,
        variable: str = None,
        location: str = None,
        time: Union[str, pd.Timestamp, int] = 0,
        ax=None,
        **kwargs,
    ):
        """Plots a profile at a single time

        Simple method to draw a profile through depth at a single timestep.
        The profile is shown on the layerfaces of the model (i.e., the vertical cell "edges")
        based on linear interpolation from the vertical cell centers.

        Args:
            variable (str, optional): variable to plot. Defaults to the first variable.
            location (str, optional): profile location to plot. Defaults to first location.
            time ([str, pd.Timestamp, int], optional): Time to plot. Defaults to 0.
            ax (plt.Axes, optional): Matplotlib axis to draw profile. Default is to create a new figure unless specified.
            kwargs (dict, optional): Keyword arguments passed to matplotlib's pyplot.plot method.

        Returns:
            np.ndarray: Array containing profile data, interpolated linearly to the layerfaces
        """

        variable, label = self._getvar_(variable, skip=["H"])
        if location is None:
            location = next(iter(self.locations.items()))[0]

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = plt.gcf()

        # Get data
        lfz = self.ts.get_raw_data("layerface_Z", location, time)[::-1, 0]
        arr = self.ts.get_raw_data(variable, location, time)[::-1, 0]
        time = self.ts.get_raw_data("Time", location, time)[0]

        z_cell = np.mean((lfz[1:], lfz[:-1]), axis=0)

        if arr.sum() == 0:
            arr_lfz = np.zeros(lfz.shape)
        else:
            arr_lfz = np.interp(lfz, z_cell, arr)

        marker = kwargs.pop("marker", ".")
        color = kwargs.pop("color", "black")
        line = ax.plot(arr_lfz, lfz, "-", marker=marker, color=color, **kwargs)

        date_title = pd.to_datetime(time).strftime("%Y-%m-%d %H:%M")

        ax.grid(True)
        ax.set_ylabel("Elevation (m)")
        ax.set_xlabel(label)
        ax.set_title(date_title)
        ax.set_xlim([arr_lfz.min() * 0.975, arr_lfz.max() * 1.025])

        return arr_lfz


def _prep_axis(ax, kwargs, equal=True):
    """
    Utility to generate new mpl fig/ax if required,
    otherwise, check if the axis requires zooming
    """
    if ax is None:
        fig, ax = plt.subplots()

        # For domain plots, etc. Curtain plots don't want this.
        if equal == True:
            ax.axis("equal")

        zoom = kwargs.pop("zoom", True)
    else:
        fig = plt.gcf()
        # Logic
        if (ax.get_xlim() == (0, 1)) & (ax.get_ylim() == (0, 1)):
            zoom = kwargs.pop("zoom", True)
        else:
            zoom = kwargs.pop("zoom", False)

    return fig, ax, zoom


def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter



"""A module defining all point time series extractor classes"""

import xarray as xr
import numpy as np
import pandas as pd
from netCDF4 import Dataset
from abc import ABC, abstractmethod
import datetime as dt
from tfv.miscellaneous import Expression
from tfv.geometry import Mesh
from typing import Union

time_slice_err = [
    "`time_limits` optional argument must be a slice function",
    "Examples:",
    "   slice(0, 10) for the first 10 timesteps",
    "   slice('2020-01-01', '2020-02-01') to slice between dates",
    "For more help, please refer to Pandas `.loc` or Xarray `.isel` or `.sel` methods",
]


class TimeSeries(ABC):
    """A base class which defines the API for all point time series based model result data extraction"""

    def __init__(self, file):
        """Initializes TimeSeries extractor object with model result file"""

        # Store file path string as attribute
        self.file = file
        self.sites = None

        # Prepare static TimeSeries attributes
        self.__prepare_file_handle__()
        self.__prepare_time_vector__()
        self.__prepare_locations__()

    @abstractmethod
    def get_raw_data(self, variable, site):
        """
        Query to extract raw time series data at a given location.

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        site : string
            Location at which to extract the time series data.

        Returns
        -------
        data : np.ndarray
            The raw time series data as 1D or 2D numpy array
        """

    @abstractmethod
    def get_mask_vector(self, site):
        """
        Query to extract an array that defines invalid model data.

        Parameters
        ----------
        site : string
            Location at which to extract the time series data.

        Returns
        -------
        mask : np.ndarray
            Logical index, True if model cell at time step is invalid (i.e dry).
        """

    @abstractmethod
    def get_z_layer_faces(self, site):
        """
        Query to extract an array that defines the vertical layer faces of a 3D model at a given location.

        Parameters
        ---------
        site : string
            Location at which to extract the time series data.

        Returns
        -------
        lfz : np.ndarray
            Vertical layer faces. If model is 2D returns None.
        """

    @abstractmethod
    def get_integral_data(self, site, datum, limits):
        """
        Query to extract data for vertical integration at a given location. Principle data is the
        integral limit (z2 - z1) and dz for each 3D model cell at the location.

        Parameters
        ----------
        site : string
            Location at which to extract the time series data.
        datum : string
            Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        limits : tuple
            Vertical depth-averaging limits relative to vertical datum.

        Returns
        -------
        (z2_z1, dz) : tuple
            The elevation limits (z2 - z1) & dz for each 3D cell at a given location
        """

    @abstractmethod
    def get_data(self, variable, site, datum="sigma", limits=(0, 1)):
        """
        Query to extract time series data at a given location. If model data is 3D then it is
        depth-averaged according to the depth-averaging vertical datum and vertical limits.

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        site : string
            Location at which to extract the time series data.
        datum : {'sigma', 'depth', 'height', 'elevation'}
            Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        limits : tuple
            Vertical depth-averaging limits (z1, z2) relative to vertical datum.
        agg : {'mean', 'min', 'max'}
            Vertical aggregration function, default = 'mean'
            'mean' will apply a weighted averaging routine (i.e., a depth average across the specified datum/limits)

        Returns
        -------
        data : np.ndarray
            1D Time series data at a given location.
        """

    @abstractmethod
    def __prepare_file_handle__(self):
        """Command which prepares the file handle for the extractor class"""

    @abstractmethod
    def __prepare_time_vector__(self):
        """Command which prepares the result time stamp vector relative to python epoch"""

    @abstractmethod
    def __prepare_locations__(self):
        """Command which prepares locations found in result time series file"""


class FvTimeSeries(TimeSeries):

    def __init__(self, file):
        super(FvTimeSeries, self).__init__(file)

    @property
    def variables(self):
        grp = list(self.locations.keys())[0]
        return [
            x for x in self.ds[grp].data_vars.keys()
            if x not in ["X", "Y", "Z", "layerface_Z", "stat"]
        ]

    @Expression.decorator
    def get_raw_data(self, variable, site, time=None):
        ds = self._subset_dataset(site, time)
        return ds[variable].values.transpose()

    def get_mask_vector(self, site, time=None):
        ds = self._subset_dataset(site, time)
        if "stat" in ds.data_vars:
            return self.get_raw_data("stat", site, time=time) == 0
        else:
            return np.zeros((ds.dims["Time"]), dtype=bool)

    def get_z_layer_faces(self, site, time=None):
        return self.get_raw_data("layerface_Z", site, time=time)

    def get_integral_data(self, site, datum, limits, time=None):
        # Get z layer faces
        lfz = self.get_z_layer_faces(site, time=time)

        # Get water level (wl) and bed level (bl) for each time step
        wli, bli = 0, lfz.shape[0] - 1
        wl, bl = lfz[wli, :], lfz[bli, :]

        # Determine integral limits z1 and z2 for each time step using wl, bl and the limits
        if datum == "sigma":
            z1 = limits[0] * (wl - bl) + bl
            z2 = limits[1] * (wl - bl) + bl
        elif datum == "height":
            z1 = limits[0] + bl
            z2 = limits[1] + bl
        elif datum == "depth":
            z1 = wl - limits[1]
            z2 = wl - limits[0]
        elif datum == "elevation":
            z1 = limits[0]
            z2 = limits[1]
        else:
            return None

        # Create integral limits, filtering z2 and z1 above and below water level or bed level
        z1 = np.minimum(np.maximum(z1, bl), wl)
        z2 = np.minimum(np.maximum(z2, bl), wl)

        # Squeeze out middle value of each vertical layer face
        lfz = np.maximum(lfz, np.tile(z1, (lfz.shape[0], 1)))
        lfz = np.minimum(lfz, np.tile(z2, (lfz.shape[0], 1)))

        # Get upper z layer face and lower z layer face for each 3D cell
        ul = np.delete(lfz, bli, axis=0)
        ll = np.delete(lfz, wli, axis=0)
        dz = ul - ll

        # Clean up integral limit to avoid division by zero
        z2_z1 = z2 - z1
        mask = z2_z1 == 0

        # Return integral limit of each 2D cell and dz of each 3D cell contained within integral limit
        return np.ma.masked_array(data=z2_z1, mask=mask, fill_value=np.nan), dz

    @Expression.decorator
    def get_data(self,
                 variable,
                 site,
                 datum="sigma",
                 limits=(0, 1),
                 agg="mean",
                 time=None):
        # Get the raw data
        data = self.get_raw_data(variable, site, time=time)
        stat = self.get_mask_vector(site, time=time)

        # Only dim names are used, so doesn't require a subset ds
        dims = self.ds[site][variable].dims

        # If data requires depth averaging
        if dims[-1] == "NumLayers":

            # Get the integral limits
            z2_z1, dz = self.get_integral_data(site, datum, limits, time=time)

            # Update stat vector with invalid limits
            stat = stat | z2_z1.mask

            if agg == "mean":
                # Integrate the data w.r.t z
                data = np.nansum(data * dz, axis=0) * (1 / z2_z1)
            elif agg == "min":
                data = np.ma.masked_array(data=data, mask=dz == 0).min(axis=0)
            elif agg == "max":
                data = np.ma.masked_array(data=data, mask=dz == 0).max(axis=0)
            else:
                assert False, "agg must be equal to 'min', 'mean' or 'max'"

        # No action required
        elif dims[-1] == "N1":
            data = data[0, :]

        # Data is 1D in z but has 2 dims (e.g. BED_MASS (time, SEDIMENT_FRACTION))
        else:
            print(
                "Warning: Data is a 1D variable with a secondary dimension: {0}"
                .format(dims[-1]))
            print("Summing data across the secondary dimension")
            print(
                "Consider using get_raw_data if this is not the desired action"
            )
            data = data.sum(axis=0)

        # Return the data
        return np.ma.masked_array(data=data, mask=stat, fill_value=np.nan)

    def get_timeseries(
            self,
            site: str,
            variables: Union[str, list] = None,
            time: slice = None,
            datum="sigma",
            limits=(0, 1),
            agg="mean",
    ):
        ds = self._subset_dataset(site, time)
        if variables is not None:
            if isinstance(variables, str):
                variables = [variables]
        else:
            variables = self.variables
        
        # Add V/VDir methods
        requested_vars = variables.copy()
        check_reqvars = False
        if "V" in variables:
            variables.extend(["V_x", "V_y"])
            variables.pop(variables.index("V"))
            check_reqvars = True
        if "VDir" in variables:
            variables.extend(["V_x", "V_y"])
            variables.pop(variables.index("VDir"))
            check_reqvars = True
        variables = np.unique(variables).tolist()

        # Prepare blank dataset with all variable info
        dropvars = [x for x in ds.data_vars if x not in variables]
        
        # Added missing_dims='ignore' for cases where profile was extracted from a 2D domain
        dst = ds.drop(["layerface_Z"] + dropvars).isel(NumLayers=0, missing_dims='ignore')
        dst.attrs[
            "Origin"] = "Timeseries extracted from TUFLOWFV cell-centered output using `tfv` python tools"
        dst.attrs["Type"] = "Timeseries cell from TUFLOWFV Output"
        dst.attrs["Datum"] = str(datum)
        dst.attrs["Limits"] = str(limits)
        dst.attrs["Agg Fn"] = agg

        for var in variables:                
            dst[var] = (
                ("Time", ),
                self.get_data(var,
                              site,
                              datum=datum,
                              limits=limits,
                              agg=agg,
                              time=time),
            )
            dst[var].attrs = ds[var].attrs

        # Very inelegant but it'll do for now. More efficient than routing through the magic expression decorator
        if "V" in requested_vars:
            dst["V"] = np.hypot(dst['V_x'], dst['V_y'])
            dst["V"].attrs = {"long_name": "current speed", "units": "m s^-1"}
        if "VDir" in requested_vars:
            dst["VDir"] = (90 - np.arctan2(dst['V_y'], dst['V_x']) * 180 / np.pi) % 360
            dst["VDir"].attrs = {"long_name": "current direction", "units": "degN TO"}

        if check_reqvars:
            if 'V_x' not in requested_vars:
                dst = dst.drop(['V_x'])
            if 'V_y' not in requested_vars:
                dst = dst.drop(['V_y'])

        return dst

    def _subset_dataset(self, site, time):
        """Helper function to subset datasets"""
        # Use sliced xarray dataset
        dsx = self.ds[site]
        int_types = [int, np.int32, np.int64, np.int16]
        date_types = [pd.Timestamp, np.datetime64, str]
        try:
            if any([isinstance(time, x) for x in int_types]):
                dsx = dsx.isel(Time=[time])
            elif isinstance(time, slice):
                if isinstance(time.stop, int):
                    dsx = dsx.isel(Time=time)
                else:
                    dsx = dsx.sel(Time=time)
            elif any([isinstance(time, x) for x in date_types]):
                dsx = dsx.sel(Time=[time], method='nearest')
            elif isinstance(time, list):
                dsx = dsx.isel(Time=time)
            elif time is None:
                dsx
            else:
                dsx = dsx.sel(Time=time)
        except AttributeError:
            print(time_slice_err)

        return dsx

    def __prepare_file_handle__(self):
        if isinstance(self.file, xr.Dataset):
            self.file = self.file.encoding["source"]
        nc = Dataset(self.file, "r")
        grps = nc.groups.keys()
        time = nc["ResTime"][:]
        tvec = pd.to_timedelta(time, unit="h") + pd.Timestamp(1990, 1, 1)

        self.ds = {}
        for grp in grps:
            ds = xr.open_mfdataset(self.file, group=grp)
            ds["Time"] = tvec
            ds.attrs["Label"] = grp
            ds.attrs["X"] = float(ds["X"][0].values)
            ds.attrs["Y"] = float(ds["Y"][0].values)
            self.ds[grp] = ds

        # TODO: Remove in future - here for legacy support
        self.nc = nc

    def __prepare_time_vector__(self):
        # Prepare time vector relative to python epoch
        fv_epoch = pd.Timestamp(1990, 1, 1)
        self.time_vector = (
                pd.to_timedelta(self.nc["ResTime"][:], unit="h") + fv_epoch
            )
        self.nt = self.time_vector.size

    def __prepare_locations__(self):
        self.locations = {}
        for group in self.nc.groups.keys():
            xp = self.nc[group]["X"][:].data[0]
            yp = self.nc[group]["Y"][:].data[0]
            self.locations[group] = [xp, yp]

    def get_geometry(self, site):
        z = self.get_z_layer_faces(site)
        t = self.time_vector
        
        # Padded time-vector for smooth meshing
        tv = pd.date_range(t[0], t[-1], periods=t.shape[0]+1)
        tc = np.tile(tv, (z.shape[0], 1))
        zc = np.hstack((z[:, [0]], 0.5 * (z[:, :-1] + z[:, 1:]), z[:, [-1]]))

        node_x, node_y = tc.ravel("F"), zc.ravel("F")

        nz, nt = tc.shape
        ii = np.arange(nz - 1)
        jj = np.arange(nt - 1)

        jj, ii = np.meshgrid(jj, ii)
        tl = np.ravel(ii + nz * jj, "F")
        tr = np.ravel(ii + nz * (jj + 1), "F")

        cell_node = np.vstack((tl, tl + 1, tr + 1, tr))
        cell_node = cell_node.transpose()

        return Mesh(node_x, node_y, cell_node)

    # Inherit doc strings (needs to be done a better way)
    get_raw_data.__doc__ = TimeSeries.get_raw_data.__doc__
    get_mask_vector.__doc__ = TimeSeries.get_mask_vector.__doc__
    get_z_layer_faces.__doc__ = TimeSeries.get_z_layer_faces.__doc__
    get_integral_data.__doc__ = TimeSeries.get_integral_data.__doc__

    get_data.__doc__ = TimeSeries.get_data.__doc__

    __prepare_file_handle__.__doc__ = TimeSeries.__prepare_file_handle__.__doc__
    __prepare_time_vector__.__doc__ = TimeSeries.__prepare_time_vector__.__doc__
    __prepare_locations__.__doc__ = TimeSeries.__prepare_locations__.__doc__



# Configuration file for the Sphinx api_reference builder.
#
# This file only contains a selection of the most common options. For a full
# list see the api_reference:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# api_reference root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath(r'../..'))

autodoc_mock_imports = ['xarray', 'netCDF4', 'tqdm', 'scipy']

# -- Project information -----------------------------------------------------

project = 'tfv'
copyright = '2023 BMT'
author = 'TUFLOW'

# The full version, including alpha/beta/rc tags
release = '1.0.1'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

extensions = \
    [
        'sphinx.ext.intersphinx',
        'sphinx.ext.autodoc',
        'sphinx.ext.napoleon',
        # 'myst_parser',
        'myst_nb',
    ]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source file names.
source_suffix = ['.rst', '.md']

# The master toc tree document.
master_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the api_reference for
# a list of builtin themes.

# html_theme = 'alabaster'
# html_theme = "sphinx_rtd_theme"

# html_theme_path = [sphinx_readable_theme.get_html_theme_path()]
# html_theme = 'readable'

html_theme = 'piccolo_theme'  # or piccolo_theme

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']


"""A module defining all objects used to visualise model result data extracts"""

# from tfv.viewer import *
from tfv.miscellaneous import *
from tfv.geometry import Mesh
from abc import ABC, abstractmethod

import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.lines import Line2D
from matplotlib.quiver import Quiver
from matplotlib.tri import Triangulation, TriContourSet
from matplotlib.cbook import silent_list
from matplotlib._tri import TriContourGenerator
from matplotlib.collections import PolyCollection, TriMesh, PathCollection

# Non-essential imports
from matplotlib import cm
from matplotlib.dates import DateFormatter
from matplotlib.colors import Normalize, BoundaryNorm


# --------------------------------------------------- Visual Objects ---------------------------------------------------
class Visual(ABC):
    """
    An abstract base class for an object which visualizes a variable. The main purpose of a Visual sub-class is to
    initialize and store a plotting library specific graphics object for a given extract type i.e SheetPatch.

    The Visual object works by connecting its set_time_current method to the Slider object. When self.set_time_current
    is called by the Slider, the Visual object selects a valid time based on the Extractor objects time_vector and updates itself.
    """

    @property
    def time_vector(self):
        return self.extractor.time_vector

    def __init__(self, axes, extractor, expression, **kwargs):
        # Set target axes
        self.axes = axes

        # Set extractor
        self.extractor = extractor
        self._time_current = self.time_vector[0]
        self._time_index = 0

        self.expression = expression
        self.expression_kwargs = dict(
            post_process=kwargs.pop("post_process", None),
        )

        # Check whether the user has supplied a normal exp
        # (e.g., str 'V_x', list of vars ['V_x', 'V_y'])
        # OR whether the user supplied a list of data [np.array, np.array]
        self.custom_data = False
        if isinstance(self.expression, list):
            if isinstance(self.expression[0], np.ndarray):
                self.__process_custom_data__()
        elif isinstance(self.expression, np.ndarray):
            self.__process_custom_data__()
        else:
            self.custom_data = False

        # Connect with viewer
        viewer = kwargs.pop("viewer", None)
        if viewer is None:
            # viewer = Viewer.__register__[0]
            self.viewer = None
        else:
            self.viewer = viewer
            self.viewer.time_vector = self.time_vector
            self.viewer.connect(self.set_time_current)

        # Initialize the graphics object
        zoom = kwargs.pop("zoom", True)
        self.__prep_graphics_obj__(**kwargs)

        # NOTE: AEW Dec 22, 2022: Moved zoom above __prep_graphics_object__
        # Unsure what ramifications this will have, but it will solve vector grid for zoomed
        # Zoom to the graphics object
        if zoom is True:
            self.zoom()

    def get_time_current(self):
        return self._time_current

    def set_time_current(self, time):
        # Add a catch to allow str or int based indexing
        if type(time) == str:
            time = pd.Timestamp(time)
        elif type(time) == int:
            time = self.time_vector[time]

        ii = np.argmin(np.abs(self.time_vector - time))

        if ii != self._time_index:
            self._time_index = ii
            self._time_current = self.time_vector[ii]
            self.__dynamic_update__()

            if self.viewer:
                self.viewer.figure.canvas.draw()

    def __process_custom_data__(self):
        self.custom_data = True
        self.vardata = np.ma.stack(self.expression)

        # Apply post-processing logic to data
        if self.expression_kwargs["post_process"]:
            fn = self.expression_kwargs["post_process"]
            self.vardata = fn(self.vardata)
        else:
            self.vardata = np.squeeze(self.vardata)

    @abstractmethod
    def __get_data__(self):
        """Abstract method which returns object data based on the expression"""

    @abstractmethod
    def __prep_graphics_obj__(self, **kwargs):
        """Abstract method to initialize the graphics object which will be used to visualize the result"""

    @abstractmethod
    def __static_update__(self):
        """Abstract method which updates static components of graphics object"""

    @abstractmethod
    def __dynamic_update__(self):
        """Abstract method which updates dynamic components graphics object"""

    @abstractmethod
    def zoom(self):
        """Abstract method which zooms axis to extent of graphics object"""


class SheetPatch(Visual):
    """
    Class for dynamic visualization of model result sheet extracts as collection of patches


    Parameters
    ----------
    axes : matplotlib.pyplot.Axes
        Axes object to display/render the graphics object
    extractor : tfv.extractor.Extractor
        Extractor object which is extracting data
    expression : string
        Expression that defines a variable
    datum : {'sigma', 'depth', 'height', 'elevation'}
        Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    limits : tuple
        Vertical depth-averaging limits (z1, z2) relative to vertical datum.
    data : numpy.array
        Custom data array of data for plotting. Should have length equal to either
        NumCells2D (if uses shading==flat, the default default) or NumVert2D if shading =='interp'


    Other Parameters
    ----------------
    shading : {'flat', 'interp'}
        Sets the shading to be flat (default) or interpolated (gourad)
    zorder : integer
        Layer order in which graphics object will be rendered (0 is bottom/first)
    cmap : matplotlib.colors.LinearSegmentedColormap
        Colormap object for mapping normalized data (0 - 1) to rgb colors
    norm : matplotlib.colors.BoundaryNorm
        Normalization object for normalizing raw data to (0 - 1) based on (min, max)
    clim : tuple
        Colour limits of the underlying matplotlib.collections.PatchCollection object (min, max)
    edgecolor : string
        Sets the mesh edge colouring of the underlying matplotlib.collections object
    antialiased : bool
        Sets the antialiasing state for rendering the underlying matplotlib.collections object
    alpha : float
        Sets the transparencies of the underlying matplotlib.collections object

    """

    def __init__(
        self,
        axes,
        extractor,
        expression,
        datum="sigma",
        limits=(0, 1),
        agg="mean",
        mask_dry=True,
        **kwargs
    ):
        self.datum = datum
        self.limits = limits
        self.agg = agg
        self.mask_dry = mask_dry

        # Call initialize method of super class
        super(SheetPatch, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (
            self.expression,
            self._time_index,
            self.datum,
            self.limits,
            self.agg,
            self.mask_dry,
        )
        if self.shading == "flat":
            return self.extractor.get_sheet_cell(*args, **self.expression_kwargs)
        elif self.shading == "interp":
            return self.extractor.get_sheet_node(*args, **self.expression_kwargs)

    def __prep_graphics_obj__(self, **kwargs):
        # Pop key word arguments which are not used by the graphics object
        self.shading = kwargs.pop("shading", "flat")

        # Get handles on sheet geometry
        node_x = self.extractor.node_x
        node_y = self.extractor.node_y
        cell_node = self.extractor.cell_node
        tri_cell_node = self.extractor.tri_cell_node

        # # Instantiate graphics object based on shading type
        if self.custom_data:
            data = self.vardata
        else:
            data = self.__get_data__()

        if self.shading == "flat":
            xy = np.dstack((node_x[cell_node], node_y[cell_node]))
            self.patch = PolyCollection(xy, array=data, **kwargs)
        elif self.shading == "interp":
            self.tri = Triangulation(node_x, node_y, triangles=tri_cell_node)
            self.patch = TriMesh(self.tri, array=data, antialiased=True, **kwargs)

            # Mask invalid triangles
            mask = np.any(data.mask[self.extractor.tri_cell_node], axis=1)
            self.tri.set_mask(mask)

        # Add the graphics object to axes
        self.axes.add_collection(self.patch)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        data = self.__get_data__()
        self.patch.set_array(data)

        # Mask invalid triangles
        if self.shading == "interp":
            mask = np.any(data.mask[self.extractor.tri_cell_node], axis=1)
            self.tri.set_mask(mask)

    def zoom(self):
        self.axes.set_xlim(
            [np.min(self.extractor.node_x), np.max(self.extractor.node_x)]
        )
        self.axes.set_ylim(
            [np.min(self.extractor.node_y), np.max(self.extractor.node_y)]
        )


class SheetContour(Visual):
    """
    Class for dynamic visualization of model result sheet extracts as contours

    Parameters
    ----------
    axes : matplotlib.pyplot.Axes
        Axes object to display/render the graphics object
    extractor : tfv.extractor.Extractor
        Extractor object which is extracting data
    expression : string
        Expression that defines a variable
    datum : {'sigma', 'depth', 'height', 'elevation'}
        Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    limits : tuple
        Vertical depth-averaging limits (z1, z2) relative to vertical datum.

    Other Parameters
    ----------------
    zorder : integer
        Layer order in which graphics object will be rendered (0 is bottom/first)
    cmap : matplotlib.colors.LinearSegmentedColormap
        Colormap object for mapping normalized data (0 - 1) to rgb colors
    norm : matplotlib.colors.BoundaryNorm
        Normalization object for normalizing raw data to (0 - 1) based on (min, max)
    clim : tuple
        Colour limits of the underlying matplotlib.collections.PatchCollection object (min, max)
    edgecolor : string
        Sets the mesh edge colouring of the underlying matplotlib.collections object
    antialiased : bool
        Sets the antialiasing state for rendering the underlying matplotlib.collections object
    alpha : float
        Sets the transparencies of the underlying matplotlib.collections object
    """

    def __init__(
        self,
        axes,
        extractor,
        expression,
        datum="sigma",
        limits=(0, 1),
        agg="mean",
        mask_dry=True,
        **kwargs
    ):
        self.datum = datum
        self.limits = limits
        self.agg = agg
        self.mask_dry = mask_dry

        # Call initialize method of super class
        super(SheetContour, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (
            self.expression,
            self._time_index,
            self.datum,
            self.limits,
            self.agg,
            self.mask_dry,
        )
        return self.extractor.get_sheet_node(*args, **self.expression_kwargs)

    def __prep_graphics_obj__(self, **kwargs):
        # Get handles on sheet geometry
        node_x = self.extractor.node_x
        node_y = self.extractor.node_y
        tri_cell_node = self.extractor.tri_cell_node

        # Get triangular mesh and initialize cpp triangulation
        tri = Triangulation(node_x, node_y, triangles=tri_cell_node)
        self.tri = tri
        self.cpp_tri = tri.get_cpp_triangulation()

        # Get data
        if self.custom_data:
            data = self.vardata
        else:
            data = self.__get_data__()

        # Mask bad triangles
        mask = np.any(data.mask[self.extractor.tri_cell_node], axis=1)
        self.cpp_tri.set_mask(mask)
                
        # Set contour limits/levels
        zlim = [data.min(), data.max()]
        
        # BUG: This is a workaround for matplotlib's TriContourSet, which has issues on a log scale with nega values
        # even if they are masked. Hence we will block out this data
        data.data[data.mask] = zlim[0]

        clim = kwargs.pop("clim", [zlim[0], zlim[1]])
        self.levels = kwargs.pop("levels", 50)

        if type(self.levels) is int:
            levels = np.linspace(clim[0], clim[1], self.levels)
        else:
            levels = self.levels

        self.cont_args = kwargs  # For dynamic update
        filled = kwargs.pop("filled", True)
        extend = kwargs.pop("extend", "both")
        self.cont = TriContourSet(
            self.axes, tri, data.data, levels, filled=filled, extend=extend, **kwargs
        )

        if clim is not None:
            self.cont.set_clim(clim)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        # Get data
        if self.custom_data:
            assert False, "You can't update time dynamically when supplying custom data"
        data = self.__get_data__()

        # Mask bad triangles
        mask = np.any(data.mask[self.extractor.tri_cell_node], axis=1)
        self.cpp_tri.set_mask(mask)

        # Set contour limits/levels
        zlim = [data.min(), data.max()]
        clim = self.cont.get_clim()

        if type(self.levels) is int:
            levels = np.linspace(clim[0], clim[1], self.levels)
        else:
            levels = self.levels

        for collection in self.cont.collections:
            self.cont.collections.remove(collection)

        self.cont = TriContourSet(
            self.axes,
            self.tri,
            data.data,
            levels,
            filled=True,
            extend="both",
            **self.cont_args
        )

        self.cont.set_clim(clim)

    def zoom(self):
        self.axes.set_xlim(
            [np.min(self.extractor.node_x), np.max(self.extractor.node_x)]
        )
        self.axes.set_ylim(
            [np.min(self.extractor.node_y), np.max(self.extractor.node_y)]
        )


class SheetVector(Visual):
    """
    Class for dynamic visualization of model result sheet extracts as gridded vector field

    Parameters
    ----------
    axes : matplotlib.pyplot.Axes
        Axes object to display/render the graphics object
    extractor : tfv.extractor.Extractor
        Extractor object which is extracting data
    expression : tuple
        Tuple of string expressions that defines vector (vec_x, vec_y)
    datum : {'sigma', 'depth', 'height', 'elevation'}
        Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    limits : tuple
        Vertical depth-averaging limits (z1, z2) relative to vertical datum.
    """

    def __init__(
        self,
        axes,
        extractor,
        expression,
        datum="sigma",
        limits=(0, 1),
        agg="mean",
        normalise=False,
        **kwargs
    ):
        self.datum = datum
        self.limits = limits
        self.agg = agg
        self.norm = normalise

        self.xg = None
        self.yg = None
        self.grid_index = None

        # Call initialize method of super class
        super(SheetVector, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (self._time_index, self.xg, self.yg, self.datum, self.limits, self.agg)
        u = self.extractor.get_sheet_grid(
            self.expression[0], *args, grid_index=self.grid_index
        )
        v = self.extractor.get_sheet_grid(
            self.expression[1], *args, grid_index=self.grid_index
        )

        return u, v

    def __prep_graphics_obj__(self, **kwargs):
        # Pop key word arguments which are not used by the graphics object
        self.resolution = kwargs.pop("resolution", 40)

        # Instantiate graphics object with no data, data will be added on draw event automatically
        self.quiver = Quiver(
            self.axes,
            np.mean(self.extractor.node_x),
            np.mean(self.extractor.node_y),
            0,
            0,
            **kwargs
        )

        # Add the graphics object to axes
        self.axes.add_collection(self.quiver)

        # Connect the regrid method to the draw event of the figure object.
        self._x_cid = self.axes.callbacks.connect(
            "xlim_changed", self.__static_update__
        )
        self._y_cid = self.axes.callbacks.connect(
            "ylim_changed", self.__static_update__
        )

    def __static_update__(self, event=None):
        xlim = self.axes.get_xlim()
        ylim = self.axes.get_ylim()

        self.xg = np.linspace(xlim[0], xlim[1], self.resolution)
        self.yg = np.linspace(ylim[0], ylim[1], self.resolution)
        self.grid_index = self.extractor.get_grid_index(self.xg, self.yg)

        u, v = self.__get_data__()
        msk = u.mask  # This is for the first timestep when Vx Vy = 0.
        if self.norm == True:
            r = np.power(np.add(np.power(u, 2), np.power(v, 2)), 0.5)
        else:
            r = 1
        u = u / r
        v = v / r

        x, y = [arr.flatten() for arr in np.meshgrid(self.xg, self.yg)]
        xy = np.hstack((x[:, np.newaxis], y[:, np.newaxis]))

        self.quiver.set_offsets(xy)
        self.quiver.N = xy.shape[0]
        self.quiver.set_UVC(u.flatten(), v.flatten())

    def __dynamic_update__(self):
        u, v = self.__get_data__()
        if self.norm == True:
            r = np.power(np.add(np.power(u, 2), np.power(v, 2)), 0.5)
        else:
            r = 1
        self.quiver.set_UVC(u / r, v / r)

    def zoom(self):
        self.axes.set_xlim(
            [np.min(self.extractor.node_x), np.max(self.extractor.node_x)]
        )
        self.axes.set_ylim(
            [np.min(self.extractor.node_y), np.max(self.extractor.node_y)]
        )


class CurtainPatch(Visual):
    """
    Class for dynamic visualization of model result curtain extracts as collection of patches.

    Arguments
    ---------
    :param axes -- object | Axes object.
    :param extractor -- object | Extractor object.
    :param expression -- string | String expression that defines variable.
    :param polyline -- 2D array | Polyline as [x, y] used to slice 3D data.

    Returns
    -------
    :param curtain -- object | CurtainPatch instance.

    Examples
    --------
    >> xtr = FvExtractor('my_file_3D.nc')
    >> curtain = CurtainPatch(axes, xtr, polyline, 'SAL')
    """

    def __init__(self, axes, extractor, expression, polyline, **kwargs):
        self.polyline = polyline
        self.x_data = extractor.get_intersection_data(polyline)
        self.index = extractor.get_curtain_cell_index(polyline)

        # Call initialize method of super class
        super(CurtainPatch, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        self.geo = self.extractor.get_curtain_cell_geo(
            self._time_index, self.polyline, self.x_data, self.index
        )

        args = (
            self.expression,
            self._time_index,
            self.polyline,
            self.x_data,
            self.index,
        )
        return self.extractor.get_curtain_cell(*args, **self.expression_kwargs)

    def __prep_graphics_obj__(self, **kwargs):
        if self.custom_data:
            data = self.vardata
        else:
            data = self.__get_data__()

        node_x, node_y, cell_node = self.geo

        xy = np.dstack((node_x[cell_node], node_y[cell_node]))
        self.patch = PolyCollection(xy, array=data, **kwargs)

        # Add the graphics object to axes
        self.axes.add_collection(self.patch)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        data = self.__get_data__()
        node_x, node_y, cell_node = self.geo

        xy = np.dstack((node_x[cell_node], node_y[cell_node]))

        self.patch.set_verts(xy)
        self.patch.set_array(data)

    def zoom(self):
        self.axes.set_xlim([np.min(self.geo[0]), np.max(self.geo[0])])
        self.axes.set_ylim([np.min(self.geo[1]), np.max(self.geo[1])])


class CurtainVector(Visual):
    """
    Class for dynamic visualization of model result curtain extracts as gridded vector field

    Arguments
    ---------
    :param axes -- object | Axes object.
    :param extractor -- object | Extractor object.
    :param expression -- string | String expression that defines variable.
    :param polyline -- 2D array | Polyline as [x, y] used to slice 3D data.

    Returns
    -------
    :param curtain -- object | CurtainPatch instance.

    Examples
    --------
    >> xtr = FvExtractor('my_file_3D.nc')
    >> curtain = SheetContour(axes, xtr, polyline, ['V', 'W'])
    """

    def __init__(
        self, axes, extractor, expression, polyline, tangential=True, **kwargs
    ):
        self.polyline = polyline
        self.x_data = extractor.get_intersection_data(polyline)
        self.index = extractor.get_curtain_cell_index(polyline)
        self.tangential = tangential

        self.xg = np.array([0, 1])
        self.yg = np.array([0, 1])

        # Call initialize method of super class
        super(CurtainVector, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        self.geo = self.extractor.get_curtain_cell_geo(
            self._time_index,
            self.polyline,
            self.x_data,
            self.index,
            return_unit_vector=True,
        )

        args = (
            self._time_index,
            self.polyline,
            self.xg,
            self.yg,
            self.x_data,
            self.index,
        )
        
        u = self.extractor.get_curtain_grid(self.expression[0], *args)
        v = self.extractor.get_curtain_grid(self.expression[1], *args)

        # Get the tangential component to the patch faces
        if self.tangential:
            grid_index = Mesh(*self.geo[:3]).get_grid_index(self.xg, self.yg)
            mask = np.equal(grid_index, -999)
            valid = np.equal(mask, False)
            uv = np.stack((u.ravel(), v.ravel()))

            utang = self.geo[4]
            utang_msk = utang[:, grid_index[valid]]
            tang = np.sum(uv[:, valid.ravel()].conj() * utang_msk, axis=0)

            u[valid] = -tang * utang_msk[0, :]

            if "W" in self.extractor.variables:
                v[valid] = self.extractor.get_curtain_grid("W", *args)[valid]
            else:
                v[valid] = 0

        return u, v

    def __prep_graphics_obj__(self, **kwargs):
        # Pop key word arguments which are not used by the graphics object
        self.resolution = kwargs.pop("resolution", 40)

        # Update geometry
        self.__get_data__()

        # Instantiate graphics object with no data, data will be added on draw event automatically
        self.quiver = Quiver(
            self.axes, np.mean(self.geo[0]), np.mean(self.geo[1]), 0, 0, **kwargs
        )

        # Add the graphics object to axes
        self.axes.add_collection(self.quiver)

        # Connect the regrid method to the draw event of the figure object.
        self._cid = self.axes.figure.canvas.mpl_connect(
            "draw_event", lambda event: self.__dynamic_update__()
        )

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        xlim = self.axes.get_xlim()
        ylim = self.axes.get_ylim()

        self.xg = np.linspace(xlim[0], xlim[1], self.resolution)
        self.yg = np.linspace(ylim[0], ylim[1], self.resolution)

        u, v = self.__get_data__()

        x, y = [arr.flatten() for arr in np.meshgrid(self.xg, self.yg)]
        xy = np.hstack((x[:, np.newaxis], y[:, np.newaxis]))

        self.quiver.set_offsets(xy)
        self.quiver.N = xy.shape[0]
        self.quiver.set_UVC(u, v)

    def zoom(self):
        self.axes.set_xlim([np.min(self.geo[0]), np.max(self.geo[0])])
        self.axes.set_ylim([np.min(self.geo[1]), np.max(self.geo[1])])


class ProfileCell(Visual):
    """
    Class for dynamic visualization of model result profile extracts as vertical line

    Arguments
    ---------
    :param axes -- object | Axes object.
    :param extractor -- object | Extractor object.
    :param expression -- string | String expression that defines variable.
    :param point -- tuple | Point (x, y) of profile location.

    Returns
    -------
    :param profile -- object | ProfileCell instance.

    Examples
    --------
    >> xtr = FvExtractor('my_file_3D.nc')
    >> profile = ProfileCell(axes, xtr, point, 'SAL')
    """

    def __init__(self, axes, extractor, expression, point, **kwargs):
        self.point = point

        super(ProfileCell, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        self.elevation = self.extractor.get_profile_cell_geo(
            self._time_index, self.point
        )

        args = (self.expression, self._time_index, self.point)
        return self.extractor.get_profile_cell(*args, **self.expression_kwargs)

    def __prep_graphics_obj__(self, **kwargs):
        if self.custom_data:
            data = self.vardata
        else:
            data = self.__get_data__()
        self.line = Line2D(data, self.elevation, **kwargs)

        self.axes.add_line(self.line)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        data = self.__get_data__()
        self.line.set_data(data, self.elevation)

    def zoom(self):
        data = self.__get_data__()
        if data.min() != 0 and data.max() != 0:
            self.axes.set_xlim([data.min() * 0.95, data.max() * 1.05])
        self.axes.set_ylim([self.elevation.min() * 0.95, self.elevation.max() * 1.05])


class SeriesGlider(Visual):
    """
    Class for dynamic visualization of model result time series extracts with current time marked

    Arguments
    ---------
    :param axes -- object | Axes object.
    :param extractor -- object | Extractor object.
    :param expression -- string | String expression that defines variable.
    :param location -- tuple | Point (x, y) of profile location.

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param profile -- object | ProfileCell instance.

    Examples
    --------
    >> xtr = FvExtractor('my_file_3D.nc')
    >> profile = ProfileCell(axes, xtr, point, 'SAL')
    """

    def __init__(
        self,
        axes,
        extractor,
        expression,
        location,
        datum="sigma",
        limits=(0, 1),
        **kwargs
    ):
        self.location = location
        self.datum = datum
        self.limits = limits

        super(SeriesGlider, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (self.expression, self.location, self.datum, self.limits)
        return self.extractor.get_data(*args)

    def __prep_graphics_obj__(self, **kwargs):
        self.date_fmt = kwargs.pop("date_fmt", "%d/%m/%Y")

        x = self.extractor.time_vector
        y = self.__get_data__()

        self.time_series = Line2D(x, y, **kwargs)
        self.axes.add_line(self.time_series)

        glider_spec = dict(linewidth=0.8, linestyle="--", color="black")
        self.glider = Line2D([x[0], x[0]], [-(10**20), 10**20], **glider_spec)
        self.axes.add_line(self.glider)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        tc = self.get_time_current()
        x = np.array([tc, tc])
        self.glider.set_xdata(x)

    def zoom(self):
        x, y = self.time_series.get_data()
        xlim = [x.min(), x.max()]
        ylim = [y.min(), y.max()]
        self.axes.set_xlim(xlim)
        self.axes.set_ylim(ylim)


class HovmollerGlider(Visual):
    def __init__(self, axes, extractor, expression, location, **kwargs):
        self.location = location

        super(HovmollerGlider, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (self.expression, self.location)
        return self.extractor.get_raw_data(*args).ravel("F")

    def __prep_graphics_obj__(self, **kwargs):
        self.mesh = self.extractor.get_geometry(self.location)
        self.mesh.node_x = mdates.date2num(pd.to_datetime(self.mesh.node_x))
        node_x = self.mesh.node_x[self.mesh.cell_node]
        node_y = self.mesh.node_y[self.mesh.cell_node]
        verts = np.dstack((node_x, node_y))

        self.patch = PolyCollection(verts, array=self.__get_data__(), **kwargs)
        self.axes.add_collection(self.patch)

        glider_spec = dict(linewidth=0.8, linestyle="--", color="black")
        self.glider = Line2D(
            self.time_vector[[0, 0]], [-(10**20), 10**20], **glider_spec
        )
        self.axes.add_line(self.glider)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        tc = self.get_time_current()
        x = np.array([tc, tc])
        self.glider.set_xdata(x)

    def zoom(self):
        xlim = [self.mesh.node_x.min(), self.mesh.node_x.max()]
        ylim = [self.mesh.node_y.min(), self.mesh.node_y.max()]
        self.axes.set_xlim(xlim)
        self.axes.set_ylim(ylim)


# --------------------------------------------- Particle Tracking Objects ----------------------------------------------
class ParticlesScatter(Visual):
    """
    Class for dynamic visualization of particle tracking model results as collection of patches

    Parameters
    ----------
    axes : matplotlib.pyplot.Axes
        Axes object to display/render the graphics object
    extractor : tfv.particles.FvParticles
        Extractor object which is extracting data
    expression : string
        Expression used to color the particles
    datum : {'sigma', 'depth', 'height', 'elevation'}
        Vertical datum applied for particle selection
    limits : tuple
        Vertical limits (z1, z2) applied for particle selection.
    show : string
        Expression used to select particles for display (works on top of vertical selection)
    highlight : string
        Expression used to highlight particles
    scale : float
        Number used to scale marker size (this is in map units i.e metres or degrees)
    shape : {'tri','square','o','star','diamond'}
        Shape of marker

    Other Parameters
    ----------------
    zorder : integer
        Layer order in which graphics object will be rendered (0 is bottom/first)
    cmap : matplotlib.colors.LinearSegmentedColormap
        Colormap object for mapping normalized data (0 - 1) to rgb colors
    norm : matplotlib.colors.BoundaryNorm
        Normalization object for normalizing raw data to (0 - 1) based on (min, max)
    clim : tuple
        Colour limits of the underlying matplotlib.collections.PatchCollection object (min, max)
    edgecolor : string
        Sets the edge colouring of the underlying matplotlib.collections object
    facecolor : string
        Sets the face colouring of the underlying matplotlib.collections object
    antialiased : bool
        Sets the antialiasing state for rendering the underlying matplotlib.collections object
    alpha : float
        Sets the transparencies of the underlying matplotlib.collections object
    """

    def __init__(
        self,
        axes,
        extractor,
        expression=None,
        datum=None,
        limits=None,
        show=None,
        highlight=None,
        track=0,
        scale=0.001,
        shape="tri",
        **kwargs
    ):

        self.datum = datum
        self.limits = limits
        self.track = track
        self.scale = scale
        self.shape = shape

        if show is not None:
            self.show = Expression(show)
        else:
            self.show = show
        if highlight is not None:
            self.highlight = Expression(highlight)
        else:
            self.highlight = highlight

        # Call initialize method of super class
        super(ParticlesScatter, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        return self.extractor.get_raw_data(self.expression, self._time_index)

    def __prep_graphics_obj__(self, **kwargs):
        # Initialize dimensionless shape
        pi = np.pi
        n = None
        r = pi / 2

        shape = self.shape.lower()
        if shape in ["triangle", "tri"]:
            n = 3
        elif shape == "square":
            n = 4
        elif shape in ["circle", "o"]:
            n = 50
        elif shape in ["diamond", "d"]:
            n = 4
            r = 0
        elif shape == "star":
            rad = np.linspace(0, 2 * pi, 6) - pi / 5 - r
            rad = rad[1, 3, 5, 2, 4, 1]
            self.shape_x = np.cos(rad)
            self.shape_y = np.sin(rad)
        else:
            pass

        if n is not None:
            rad = np.linspace(0, 2 * pi, n + 1) - pi / n - r
            self.shape_x = np.cos(rad)
            self.shape_y = np.sin(rad)

        node_x = 0 + self.scale * self.shape_x
        node_y = 0 + self.scale * self.shape_y

        xy = np.dstack((node_x, node_y))
        self.patch = PolyCollection(xy, **kwargs)

        # Add the graphics object to axes
        self.axes.add_collection(self.patch)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        # Get logical index of particles to show based on self.show expression
        if self.show is None:
            show_lgi = np.ones((self.extractor.np,), dtype=np.bool)
        else:
            show_lgi = self.extractor.get_raw_data(self.show, self._time_index)

        # Apply additional elevation & inactivity filters
        if self.datum is None and self.limits is None:
            # Get valid particles based on mask only
            valid_lgi = np.invert(self.extractor.get_mask_vector(self._time_index))
        else:
            # Get valid particles based elevation filter (mask applied in get_vertical selection)
            valid_lgi = self.extractor.get_vertical_selection(
                self._time_index, self.datum, self.limits
            )

        # Update the 'show' logical index
        show_lgi = show_lgi & valid_lgi

        # Count number to show
        ns = np.sum(show_lgi)

        # Plot the particles
        if ns != 0:
            # Get logical index of particles to highlight
            if self.highlight is None:
                # Default to highlighting none of the shown particles
                highlight_lgi = np.zeros((ns,), dtype=np.bool)
            else:
                # Get highlight logical index (sub setting with the 'show' lgi)
                highlight_lgi = self.extractor.get_raw_data(
                    self.highlight, self._time_index
                )[show_lgi]
            highlight_lgi[0] = False  # Stops from highlighting all particles (bug)
            nh = np.sum(highlight_lgi)  # Count number to highlight

            # Update patch faces (sub setting with the 'show' lgi)
            x = self.extractor.get_raw_data("x", self._time_index)[show_lgi]
            y = self.extractor.get_raw_data("y", self._time_index)[show_lgi]

            node_x = x.reshape(x.size, 1) + self.scale * self.shape_x
            node_y = y.reshape(y.size, 1) + self.scale * self.shape_y

            xy = np.dstack((node_x, node_y))
            self.patch.set_verts(xy)

            # Update colours
            if self.expression is None:
                # Get face colours
                c = self.patch.get_facecolor()[0]
                face_colours = np.tile(c, (ns, 1))

                # Finish highlighting with yellow
                y = np.array([255, 255, 0.0, 255.0]) / 255
                face_colours[highlight_lgi, :] = np.tile(y, (nh, 1))

                # Set face colours
                self.patch.set_facecolor(face_colours)

                # Get edge colours & set edge colors
                c = self.patch.get_edgecolor()
                if c.size > 0:
                    # Set edge colours
                    edge_colours = np.tile(c[0], (ns, 1))
                    self.patch.set_edgecolor(edge_colours)
            else:
                # Get data relating to self.expression (sub setting with the 'show' lgi)
                data = self.extractor.get_raw_data(self.expression, self._time_index)[
                    show_lgi
                ]

                # Get face colours
                norm = self.patch.norm(data)
                face_colours = self.patch.cmap(norm)

                # Finish highlighting
                y = np.array([255, 255, 0.0, 255]) / 255
                face_colours[highlight_lgi, :] = np.tile(y, (nh, 1))

                # Set face colours
                self.patch.set_facecolor(face_colours)

                # Get & set edge colours
                c = self.patch.get_edgecolor()
                if c.size > 0:
                    # Set edge colours
                    edge_colours = np.tile(c[0], (ns, 1))
                    self.patch.set_edgecolor(edge_colours)
        else:
            node_x = np.nan + self.scale * self.shape_x
            node_y = np.nan + self.scale * self.shape_y

            xy = np.dstack((node_x, node_y))
            self.patch.set_verts(xy)

    def zoom(self):
        pass


class ParticlesHeat(Visual):
    def __init__(
        self, axes, extractor, expression, datum="sigma", limits=(0, 1), **kwargs
    ):
        self.datum = datum
        self.limits = limits

        # Call initialize method of super class
        super(ParticlesHeat, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (self._time_index, self.datum, self.limits, self.count)

        return self.extractor.get_particle_density(*args)

    def __prep_graphics_obj__(self, **kwargs):
        # Pop key word arguments which are not used by the graphics object
        self.count = kwargs.pop("count", 10000)
        self.index = np.arange(
            0, self.extractor.np, np.ceil(self.extractor.np / self.count)
        ).astype(dtype=np.int32)

        x = self.extractor.get_particle_data("x", self._time_index)[self.index]
        y = self.extractor.get_particle_data("y", self._time_index)[self.index]

        # Instantiate graphics object with no data, data will be added on draw event automatically
        self.sca = self.axes.scatter(x, y, **kwargs)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        x = self.extractor.get_particle_data("x", self._time_index)[self.index]
        y = self.extractor.get_particle_data("y", self._time_index)[self.index]
        data = self.__get_data__()

        self.sca.set_offsets(np.column_stack((x, y)))
        self.sca.set_array(data)

    def zoom(self):
        pass


# ------------------------------------------------ Delta Visual Objects ------------------------------------------------
class DeltaSheetPatch(SheetPatch):
    """
    Class for dynamic visualization of model result sheet extract differences as collection of patches

    Arguments
    ---------
    :param axes -- object | Axes object
    :param extractor -- tuple | Extractor objects (base, developed)
    :param expression -- string | Expression that defines variable

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param sheet -- object | DeltaSheetPatch instance

    Examples
    --------
    >> base_xtr = FvExtractor('my_file_base.nc')
    >> dev_xtr = FvExtractor('my_file_developed.nc')
    >> diff_sheet = DeltaSheetPatch(axes, (base_xtr, dev_xtr), 'H')
    """

    def __init__(
        self, axes, extractors, expression, datum="sigma", limits=(0, 1), **kwargs
    ):
        # Store multiple extractor objects
        self.extractors = extractors

        # Initialize as super class
        super(DeltaSheetPatch, self).__init__(
            axes, self.extractors[0], expression, datum, limits, **kwargs
        )

    def __get_data__(self):
        self.extractor = self.extractors[0]
        data_1 = super(DeltaSheetPatch, self).__get_data__()

        self.extractor = self.extractors[1]
        data_2 = super(DeltaSheetPatch, self).__get_data__()

        return data_2 - data_1


class DeltaSheetContour(SheetContour):
    """
    Class for dynamic visualization of model result sheet extract differences as contours

    Arguments
    ---------
    :param axes -- object | Axes object
    :param extractor -- tuple | Extractor objects (base, developed)
    :param expression -- string | Expression that defines variable

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param sheet -- object | DeltaSheetContour instance

    Examples
    --------
    >> base_xtr = FvExtractor('my_file_base.nc')
    >> dev_xtr = FvExtractor('my_file_developed.nc')
    >> diff_sheet = DeltaSheetContour(axes, (base_xtr, dev_xtr), 'H')
    """

    def __init__(
        self, axes, extractors, expression, datum="sigma", limits=(0, 1), **kwargs
    ):
        # Store multiple extractor objects
        self.extractors = extractors

        # Initialize as super class
        super(DeltaSheetContour, self).__init__(
            axes, self.extractors[0], expression, datum, limits, **kwargs
        )

    def __get_data__(self):
        self.extractor = self.extractors[0]
        data_1 = super(DeltaSheetContour, self).__get_data__()

        self.extractor = self.extractors[1]
        data_2 = super(DeltaSheetContour, self).__get_data__()

        return data_2 - data_1


class DeltaSheetVector(SheetVector):
    """
    Class for dynamic visualization of model result sheet extract differences as gridded vector field

    Arguments
    ---------
    :param axes -- object | Axes object
    :param extractor -- tuple | Extractor objects (base, developed)
    :param expression -- string | Expression that defines variable

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param vector -- object | DeltaSheetVector instance

    Examples
    --------
    >> base_xtr = FvExtractor('my_file_base.nc')
    >> dev_xtr = FvExtractor('my_file_developed.nc')
    >> diff_sheet = DeltaSheetVector(axes, (base_xtr, dev_xtr), ['V_x', 'V_y'])
    """

    def __init__(
        self, axes, extractors, expression, datum="sigma", limits=(0, 1), **kwargs
    ):
        # Store multiple extractor objects
        self.extractors = extractors

        # Initialize as super class
        super(DeltaSheetVector, self).__init__(
            axes, self.extractors[0], expression, datum, limits, **kwargs
        )

    def __get_data__(self):
        self.extractor = self.extractors[0]
        data_1 = super(DeltaSheetVector, self).__get_data__()

        self.extractor = self.extractors[1]
        data_2 = super(DeltaSheetVector, self).__get_data__()

        return data_2[0] - data_1[0], data_2[1] - data_1[1]



from pickle import TRUE
import pkg_resources

__version__ = "1.0.1"
__author__ = "support@tulow.com"
__aus_date__ = "%d/%m/%Y %H:%M:%S"

# List of dependencies copied from requirements.txt
dependencies = \
    [
        'numpy>=1.19.0',
        'matplotlib>=3.2.2',
        'netCDF4>=1.5.3',
        'xarray>=0.16.2',
        'dask>=2021.3.0',
        'scipy>=1.6.0',
        'tqdm>=4.50.0',
    ]

# Throw exception if correct dependencies are not met
pkg_resources.require(dependencies)



"""A module defining all Particle classes. Particles are the primary objects for ptm data extraction"""

import numpy as np
import datetime as dt
from abc import ABC, abstractmethod
from netCDF4 import Dataset

from tfv.mldatetime import *
from tfv.miscellaneous import Expression


class Particles(ABC):

    def __init__(self, file):
        """Initializes Extractor object with a ptm results file i.e A TUFLOW FV netCDF4 results file."""

        # Store file path string as attribute
        self.file = file

        # Prepare static Extractor attributes
        self.__prep_file_handle__()
        self.__prep_time_vector__()
        self.__prep_2d_geometry__()
        self.__prep_3d_geometry__()

    @abstractmethod
    def get_raw_data(self, variable, ii):
        """
        Query to extract raw data at a time step (if time-varying).

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        ii : integer
            The time vector index at which to extract the data.

        Returns
        -------
        data : np.ndarray
            The raw data as 1D or 2D numpy array
        """
        pass

    @abstractmethod
    def get_mask_vector(self, ii):
        """
        Query to extract an array that defines invalid model data.

        Parameters
        ----------
        ii : integer
            Time index at which to extract the mask array.

        Returns
        -------
        mask : np.ndarray
            Logical index, True if model cells/nodes are invalid (i.e dry cells).

        """
        pass

    @abstractmethod
    def get_vertical_selection(self, ii, datum, limits):
        """
        Query to extract logical index of particles within a given vertical selection at given time step.

        Parameters
        ----------
        ii : integer
            Time index at which to extract the selection.
        datum : {'sigma', 'depth', 'height', 'elevation'}
            Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        limits : tuple
            Vertical depth-averaging limits (z1, z2) relative to vertical datum.

        Returns
        -------
        lgi : np.ndarray
            A logical index for particles within specified limits. True if particle is in limits.
        """
        pass

    @abstractmethod
    def __prep_file_handle__(self):
        """Command which prepares the file handle for the extractor class"""

    @abstractmethod
    def __prep_time_vector__(self):
        """Command which prepares the result time stamp vector relative to python epoch"""

    @abstractmethod
    def __prep_2d_geometry__(self):
        """Command which prepares the result 2D mesh geometry"""

    @abstractmethod
    def __prep_3d_geometry__(self):
        """Command which prepares the result 2D mesh geometry"""


class FvParticles(Particles):
    """
        Class that extracts particle data from a TUFLOW FV PTM netCDF4 result file.

        Parameters
        ----------
        file : string
            Model result file path.

        Attributes
        ----------
        nc : netCDF4.Dataset
            Dataset object
        nt : int
            Number of time steps
        np : int
            Number of particles
        """

    def __init__(self, file):
        super(FvParticles, self).__init__(file)

    @Expression.decorator
    def get_raw_data(self, variable, ii):
        if self.nc[variable].ndim > 1:
            return self.nc[variable][ii, :].data
        else:
            return self.nc[variable][:].data

    def get_mask_vector(self, ii):
        return self.get_raw_data('stat', ii) < 0

    def get_vertical_selection(self, ii, datum, limits):
        # Get raw data
        pz = self.get_raw_data('z', ii)
        pd = self.get_raw_data('depth', ii)
        wd = self.get_raw_data('water_depth', ii)

        # Get the mask based on inactive particles
        mask = self.get_mask_vector(ii)

        # Get water level (wl) and bed level (bl) for each particle
        wl = np.ma.masked_array(pz + pd, mask=mask, fill_value=-999)
        bl = np.ma.masked_array(wl - wd, mask=mask, fill_value=-999)

        # Convert the limits into elevation
        if datum == 'sigma':
            z1 = limits[0] * (wl - bl) + bl
            z2 = limits[1] * (wl - bl) + bl
        elif datum == 'height':
            z1 = limits[0] + bl
            z2 = limits[1] + bl
        elif datum == 'depth':
            z1 = wl - limits[1]
            z2 = wl - limits[0]
        elif datum == 'elevation':
            z1 = limits[0]
            z2 = limits[1]
        else:
            return None

        # Return logical index of cells in limits
        return (z1.filled() <= pz) & (pz <= z2.filled())

    def __prep_file_handle__(self):
        # Store netCDF4 Dataset handle
        self.nc = Dataset(self.file, 'r')

    def __prep_time_vector__(self):
        # Define fv epoch relative to python epoch
        fv_epoch = dt.datetime(1990, 1, 1)
        fv_epoch = datenum(fv_epoch)

        # Prepare time vector relative to python epoch
        fv_time_vector = self.nc['Time'][:].data
        py_time_vector = fv_time_vector*3600 + fv_epoch

        # Store data as attributes
        self.time_vector = py_time_vector
        self.nt = self.time_vector.size

    def __prep_2d_geometry__(self):
        self.np = self.nc['groupID'].size

    def __prep_3d_geometry__(self):
        pass

    # Inherit doc strings (needs to be done a better way with decorator as per matplotlib)
    get_raw_data.__doc__ = Particles.get_raw_data.__doc__
    get_mask_vector.__doc__ = Particles.get_mask_vector.__doc__
    get_vertical_selection.__doc__ = Particles.get_vertical_selection.__doc__
